/* This is the device support for the memory mapped VME/vxWorks and Linux cards */
/* The driver code was writte by M.Dach 2005 - 2008                                    */

#ifndef linux
#include <vxWorks.h>
#include <types.h>
#include <stdioLib.h>
#else
#include <stdio.h>
#include "epicsExport.h"
#endif

#include <stdlib.h>

#include <string.h>
#include <recGbl.h>
#include <alarm.h>
#include <dbDefs.h>
#include <dbAccess.h>
#include <recSup.h>
#include <devSup.h>
/* #include <dbScan.h> */
#include <aiRecord.h>
#include <aoRecord.h>
#include <biRecord.h>
#include <boRecord.h>
#include <mbbiRecord.h>
#include <mbboRecord.h>
#include <mbboDirectRecord.h>
#include <mbbiDirectRecord.h>
#include <stringinRecord.h>
#include <waveformRecord.h>
#include <longinRecord.h>
#include <longoutRecord.h>



#include <drvGPMM.h>
#include <userGPMM.h>

#define PARAM_VLIDITY(MASK,LLIM,HLIM,DEFAULT) MASK<LLIM ? DEFAULT : (MASK>HLIM ? DEFAULT : MASK)

/* ------------------ Global varaibales ---------------------------------------------------------*/
#ifdef linux
struct sIrqGpmmSup irqGpmmSup = {{NULL},{0},0,-1};
#endif

/*----------------------------------------------------------------------------*/
int getRecordParam(char *sParam, struct recDesc * pRecDesc, int *iInitRead){

  char *pcTerminator=NULL;
  char *pcTerminator2=NULL;
  char *pcTerminator3=NULL;
  char *pCheck;
  char mySeparator;
  char *myStr=NULL;
  char *myStrEnd=NULL;

  char pStr[100];
  /* initialisation of some fields */

  *iInitRead=0;
  pRecDesc->uiMask=0;
  pRecDesc->uiBytesNo=0;
  pRecDesc->cShiftBits=0;
  pRecDesc->cMaskBitNr=0;
  pRecDesc->uiParam1=0;
  pRecDesc->iIntrNr=-1;
  pRecDesc->lInArg1=-1;
  pRecDesc->lInArg2=-1;
  pRecDesc->vOutArg=NULL;

  sprintf(pStr,"%.70s",sParam);
  pcTerminator=strchr(pStr,':');
  if(pcTerminator!=NULL){  /* first occurence of : */
    pcTerminator=&pcTerminator[1];
    pcTerminator3=strchr(pcTerminator,':'); /* search for the next terminator : or , */
    if(pcTerminator3==NULL)
      pcTerminator2=strchr(pcTerminator,',');
    else
      pcTerminator2=NULL;
    
    if(pcTerminator2!=NULL || pcTerminator3!=NULL){ /* 2nd terminator was found : or , */

      if(pcTerminator3!=NULL)
	pcTerminator3[0]='\0';

      if(pcTerminator2!=NULL)
	pcTerminator2[0]='\0';

	 pRecDesc->lInArg1 = strtol(pcTerminator,&pCheck,10);
	 if(!(*pcTerminator !='\0' && *pCheck=='\0'))
	     pRecDesc->lInArg1=-1;

	 if(pcTerminator3!= NULL){ /* search for the lInArg2 */

	   pcTerminator=&pcTerminator3[1];

	   pcTerminator2=strchr(pcTerminator,',');
	   if(pcTerminator2!=NULL){ /* the , terminator was found */
	     pcTerminator2[0]='\0';
	     
	     pRecDesc->lInArg2 = strtol(pcTerminator,&pCheck,10);
	     if(!(*pcTerminator!='\0' && *pCheck=='\0'))
	       pRecDesc->lInArg2=-1;

	     pcTerminator2[0]=',';
	     pcTerminator=pcTerminator2;    /* exit form here , pcTerminator holds the pointer to the 1st occurance of , */
	   }
	   else{ /* no terminator was found */
	     pRecDesc->lInArg2 = strtol(pcTerminator,&pCheck,10);
	     if(!(*pcTerminator!='\0' && *pCheck=='\0'))
	       pRecDesc->lInArg2=-1;
	     
	     pcTerminator=NULL;            /* exit form here . There are no arguments left */
	   }
	 } /* there is only one lInArg1 */
	 else{
	   pcTerminator2[0]=',';
	   pcTerminator=pcTerminator2;     /* exit from here. There are some user options left after first , */
	 }
    }
    else{
      pRecDesc->lInArg1 = strtol(pcTerminator,&pCheck,10); /* ther is only one in Arg and there is nothing else */
      if(!(*pcTerminator!='\0' && *pCheck=='\0'))
	pRecDesc->lInArg1=-1;  
      pcTerminator=NULL;                      /* There is now  , left */
    }
  }
  else /* No : terminator was found */
    pcTerminator=strchr(pStr,',');
  


  while(pcTerminator!=NULL){
    pcTerminator2=strchr(&pcTerminator[1],',');

    if(pcTerminator2!=NULL)
      pcTerminator2[0]='\0';
    
    if(pcTerminator[1]=='m'){                      /* number of mask bits. AI/AO */
      pRecDesc->uiMask= atoi(&(pcTerminator[3]));
      
	pRecDesc->uiBytesNo=pRecDesc->uiMask>>3; /* devided by 8 */
	if(pRecDesc->uiMask % 8 > 0)
	  pRecDesc->uiBytesNo++;
      
    }
    else if(pcTerminator[1]=='i'){  
      if(!strncmp(&(pcTerminator[1]),"iRead=",6)){
	if(pcTerminator[7]=='y' || pcTerminator[7]=='Y')
	  *iInitRead=INIT_FROM_HARDWARE;
        else if (pcTerminator[7]=='v' || pcTerminator[7]=='V') /* namely VAL */
	  *iInitRead=INIT_RESTORE_VAL;
	  }
    }
    else if(pcTerminator[1]=='O'){   /* OUT */
      if(!strncmp(&(pcTerminator[1]),"OUT",3)){
	pRecDesc->uiParam1 = pRecDesc->uiParam1 | WF_REC_OUT;
      }
    }
    else if(pcTerminator[1]=='S'){   /* SEQ */
      if(!strncmp(&(pcTerminator[1]),"SEQ",3)){
	pRecDesc->uiParam1 = pRecDesc->uiParam1 | WF_SEQ;
      }
    }
    else if (pcTerminator[1]=='I'){   /* IRQ  */
      if(!strncmp(&(pcTerminator[1]),"IRQ=",4)){
		  myStr=&(pcTerminator[5]);
		  pRecDesc->iIntrNr=strtol(myStr, &myStrEnd, 0);
		  if(pRecDesc->iIntrNr==0  && myStr==myStrEnd)
		    pRecDesc->iIntrNr=-1;
      }
    }
    else if(pcTerminator[1]=='B'){                 /* number of bytes.  AI/AO/StrIn  */
      pRecDesc->uiBytesNo= atoi(&(pcTerminator[3]));
      if(pRecDesc->uiBytesNo<=4)
	pRecDesc->uiMask=(pRecDesc->uiBytesNo)<<3; /* multiplied by 8 */
    }
    else if(pcTerminator[1]=='s'){                 /* shift number of bits. BI/BO/Mbbi/Mbbo */
      myStr=&(pcTerminator[3]);
      pRecDesc->cShiftBits = strtol(myStr, &myStr, 0);
      mySeparator = *myStr++;
      /* Handle any number of optional +o additions to the st_offs */
      while (mySeparator == '+')
	{
	  pRecDesc->cShiftBits += strtol(myStr, &myStr, 0);
	  mySeparator = *myStr++;
	}
      /* It was a previous implementation */
      /* pRecDesc->cShiftBits= atoi(&(pcTerminator[3])); */
    }
      
    pcTerminator=pcTerminator2;

  } /* end of pcTerminator while */
  pRecDesc->cMaskBitNr=pRecDesc->uiMask;
  if(pRecDesc->cMaskBitNr>=32)
    pRecDesc->uiMask=0xFFFFFFFF;
  else
  pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);

  /* printf("sMask 0x%X, Bytes nr %d\n",pRecDesc->uiMask,pRecDesc->uiBytesNo); */
  return(Ok);
}



/**********************/
/*       AI          */
/*********************/

struct ai_dev_sup {
	long            number;
	DEVSUPFUN       dev_report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       read_ai;
   	DEVSUPFUN       special_linconv; 
};

STATIC long init_record_ai();
STATIC long get_ioint_info_ai();
STATIC long read_ai();
STATIC long linconv_ai();

/* Differential and Single-Ended device support */
struct ai_dev_sup devAiGPMM = {
	6,
	NULL,
	NULL,
	init_record_ai,
	get_ioint_info_ai,
	read_ai,
	linconv_ai,
};

#ifdef linux 
epicsExportAddress(dset,devAiGPMM);
#endif

/**********************/
/*       AO          */
/*********************/

struct ao_dev_sup {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       write_ao;
        DEVSUPFUN       special_linconv;
};

STATIC long init_record_ao();
STATIC long get_ioint_info_ao();
STATIC long write_ao();
STATIC long linconv_ao();

struct ao_dev_sup devAoGPMM = {
        6,
        NULL,
        NULL,
        init_record_ao,
        get_ioint_info_ao,
        write_ao,
        linconv_ao
};

#ifdef linux 
epicsExportAddress(dset,devAoGPMM);
#endif

/**********************/
/*   BI               */
/**********************/

struct bi_dev_sup {
	long            number;
	DEVSUPFUN       report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       read_bi;
};

STATIC long get_ioint_bi_info();
STATIC long init_bi_record();
STATIC long read_bi();

struct bi_dev_sup devBiGPMM = {
	5,
	NULL,
	NULL,
	init_bi_record,
	get_ioint_bi_info,
	read_bi
};
#ifdef linux 
epicsExportAddress(dset,devBiGPMM);
#endif
/**********************/
/*   BO               */
/**********************/

struct bo_dev_sup {
	long            number;
	DEVSUPFUN       report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       write_bo;
};

STATIC long get_ioint_bo_info();
STATIC long init_bo_record();
STATIC long write_bo();

struct bo_dev_sup devBoGPMM = {
	5,
	NULL,
	NULL,
	init_bo_record,
	get_ioint_bo_info,
	write_bo
};

#ifdef linux 
epicsExportAddress(dset,devBoGPMM);
#endif

/**********************/
/*   MBBI DIRECT      */
/**********************/

struct mbbiDirect_dev_sup {
	long            number;
	DEVSUPFUN       report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       read_mbbi;
};
STATIC long get_ioint_mbbiDirect_info();
STATIC long init_mbbiDirect_record();
STATIC long read_mbbiDirect();

struct mbbiDirect_dev_sup devMbbiDirectGPMM = {
	5,
	NULL,
	NULL,
	init_mbbiDirect_record,
	get_ioint_mbbiDirect_info,
	read_mbbiDirect
};
#ifdef linux 
epicsExportAddress(dset,devMbbiDirectGPMM);
#endif
/**********************/
/*   MBBO DIRECT      */
/**********************/


struct mbboDirect_dev_sup {
	long            number;
	DEVSUPFUN       dev_report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       write_mbbo;
};

STATIC long get_ioint_mbboDirect_info();
STATIC long init_mbboDirect_record();
STATIC long write_mbboDirect();

struct mbboDirect_dev_sup devMbboDirectGPMM = {
	5,
	NULL,
	NULL,
	init_mbboDirect_record,
	get_ioint_mbboDirect_info,
	write_mbboDirect
};

#ifdef linux 
epicsExportAddress(dset,devMbboDirectGPMM);
#endif
/**********************/
/*   STRING IN        */
/**********************/

struct strIn_dev_sup {
	long            number;
	DEVSUPFUN       report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       read_strIn;
};

STATIC long get_ioint_strIn_info();
STATIC long init_strIn_record();
STATIC long read_strIn();

struct strIn_dev_sup devStrInGPMM = {
	5,
	NULL,
	NULL,
	init_strIn_record,
	get_ioint_strIn_info,
	read_strIn
};
#ifdef linux 
epicsExportAddress(dset,devStrInGPMM);
#endif
/**********************/
/*   WAVEFORM        */
/**********************/

STATIC long get_ioint_wf_info();
STATIC long init_wf_record();
STATIC long read_wf();


struct {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       read_wf;
} devWfGPMM={
        5,
        NULL,
        NULL,
        init_wf_record,
        get_ioint_wf_info,
        read_wf};


#ifdef linux 
epicsExportAddress(dset,devWfGPMM);
#endif

/**********************/
/*   LONGIN            */
/**********************/

struct longIn_dev_sup {
	long            number;
	DEVSUPFUN       report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       read_longin;
};
STATIC long get_ioint_longIn_info();
STATIC long init_longIn_record();
STATIC long read_longIn();

struct longIn_dev_sup devLongInGPMM = {
	5,
	NULL,
	NULL,
	init_longIn_record,
	get_ioint_longIn_info,
	read_longIn
};
#ifdef linux 
epicsExportAddress(dset,devLongInGPMM);
#endif

/**********************/
/*       LONGOUT      */
/*********************/

struct longOut_dev_sup {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       write_longout;

};

STATIC long init_record_longOut();
STATIC long get_ioint_info_longOut();
STATIC long write_longOut();


struct longOut_dev_sup devLongOutGPMM = {
        5,
        NULL,
        NULL,
        init_record_longOut,
        get_ioint_info_longOut,
        write_longOut,
};

#ifdef linux 
epicsExportAddress(dset,devLongOutGPMM);
#endif

/**********************/
/*   MBBI       */
/**********************/

struct mbbi_dev_sup {
	long            number;
	DEVSUPFUN       report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       read_mbbi;
};
STATIC long get_ioint_mbbi_info();
STATIC long init_mbbi_record();
STATIC long read_mbbi();

struct mbbi_dev_sup devMbbiGPMM = {
	5,
	NULL,
	NULL,
	init_mbbi_record,
	get_ioint_mbbi_info,
	read_mbbi
};
#ifdef linux 
epicsExportAddress(dset,devMbbiGPMM);
#endif
/**********************/
/*   MBBO             */
/**********************/


struct mbbo_dev_sup {
	long            number;
	DEVSUPFUN       dev_report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record;
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       write_mbbo;
};

STATIC long get_ioint_mbbo_info();
STATIC long init_mbbo_record();
STATIC long write_mbbo();

struct mbbo_dev_sup devMbboGPMM = {
	5,
	NULL,
	NULL,
	init_mbbo_record,
	get_ioint_mbbo_info,
	write_mbbo
};

#ifdef linux 
epicsExportAddress(dset,devMbboGPMM);
#endif
/*---------------------------------------------------------------------------*/
/*                       Functions implementation                            */
/*---------------------------------------------------------------------------*/

/**********************/
/*       AI          */
/*********************/


STATIC long get_ioint_info_ai(int cmd,struct aiRecord *pai, IOSCANPVT *ppvt){


#ifndef linux  
  struct io_GPMM	  *descrp=NULL;
  struct register_Node    *regNode=NULL;
#else
  int iX=0;
#endif      

  if( pai->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pai->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */
    iX=((struct recDesc *)pai->dpvt)->iIntrNr;
    if( iX >= 0 )
      if(irqGpmmSup.pioScanPvt[iX])
	 *ppvt = irqGpmmSup.pioScanPvt[iX];
#endif /* linux                               */
  }

  return (0);

}

/*---------------------------------------*/

STATIC long init_record_ai(
	struct aiRecord *pai
)
{
  struct vmeio *pvmeio;
  struct recDesc  *pRecDesc;
  int iInitRead=0;
  pvmeio = (struct vmeio *)&(pai->inp.value.vmeio);

 /* ai.inp must be an VME_IO */
  switch (pai->inp.type) {
	case (VME_IO):
		break;
	default:
		recGblRecordError(S_db_badField, (void *) pai,
			"devAiGPMM (init_record) Illegal INP field");
		return(S_db_badField);
  }
  
   if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("AI record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pai->dpvt=NULL;
	pai->eslo = (pai->eguf - pai->egul) / 0xffffffff;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("AI record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */
        getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,32,32);
	pRecDesc->uiBytesNo=PARAM_VLIDITY (pRecDesc->uiBytesNo,1,4,4);
	
	if(pRecDesc->cMaskBitNr>=32)   	
	  pRecDesc->uiMask=0xFFFFFFFF; 
	else                           
	  pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	 
	pai->dpvt=(char *)pRecDesc; 
	pai->eslo = (pai->eguf - pai->egul) / pRecDesc->uiMask;
      }
      
	
      
    }
   else
     pai->eslo = (pai->eguf - pai->egul) / 0xffffffff;
  
  /* set linear conversion slope */
  

  pai->udf = FALSE; /* Bob said that :reason already known */

  return(Ok);
}

/*---------------------------*/

STATIC long read_ai(struct aiRecord *pai)
{
  struct vmeio *pvmeio;/*  = &pai->inp.value.vmeio; */
  int status=Ok;
  unsigned int iVal=0;

  pvmeio = (struct vmeio *)&(pai->inp.value.vmeio);

  status = READ_AI_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pai->dpvt,REC_AI,pai->name, &iVal);
  /* printf("Ai Val = %d Mask %d \n",iVal,((struct recDesc *)(pvmeio->dpvt))->uiMask); */
  if( status == Ok){
    iVal=iVal & ((struct recDesc *)(pai->dpvt))->uiMask;
    pai->rval=iVal;
  }
 
 /* rval and val are already set in the function */

	if (status != Ok) {
	
		recGblSetSevr(pai, READ_ALARM, INVALID_ALARM);
	}
	else{
	/* linera conversion */ 
	  if(pai->linr == 1) 
	    pai->val = pai->eslo * pai->rval + pai->egul;
	  else
	    pai->val=iVal;
	}
	
	return(2); /* 2 means no convertion */
}

/*---------------------------*/

STATIC long linconv_ai(
	struct aiRecord *pai,
	int after
)
{
  
  /* if (!after)
     return(Ok);  */
  	return(Ok); 
   pai->eslo = ((float) (pai->eguf - pai->egul)) / (float) 0xffffffff; 
  
  /* printf("convertion routine called \n"); */

  return(Ok);
}

/**********************/
/*      AO            */
/*********************/

STATIC long get_ioint_info_ao(int cmd,struct aoRecord *pao, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else
  int iX=0;
#endif

  if( pao->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pao->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */
    iX=((struct recDesc *)pao->dpvt)->iIntrNr;
    if( iX >= 0 )
      if(irqGpmmSup.pioScanPvt[iX])
	 *ppvt = irqGpmmSup.pioScanPvt[iX];
#endif /* linux */
  }
  return (0);

}

/*------------------------------------*/

STATIC long init_record_ao(struct aoRecord *pao)
{
  /* int         status=Ok; */
    struct aiRecord pai;
    struct recDesc  *pRecDesc=NULL;
    int iInitRead=0;
    struct vmeio *pvmeio = &pao->out.value.vmeio;

 /* ao.out must be an VME_IO */
    if (pao->out.type != VME_IO) {
  
		recGblRecordError(S_db_badField, (void *) pao,
			"devAoGPMM (init_record) Illegal OUT field");
		return(S_db_badField);
  }

    /* copy region begins here */
    if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("AO record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pao->dpvt=NULL;
	pao->eslo = (pao->eguf - pao->egul) / 0xffffffff;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("AO record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */
        getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,32,32);
	pRecDesc->uiBytesNo=PARAM_VLIDITY (pRecDesc->uiBytesNo,1,4,4);

	if(pRecDesc->cMaskBitNr>=32)   
	  pRecDesc->uiMask=0xFFFFFFFF; 
	else                           	
	  pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	 
	pao->dpvt=(char *)pRecDesc; 
	pao->eslo = (pao->eguf - pao->egul) / pRecDesc->uiMask;
      }
      
    }
   else
     pao->eslo = (pao->eguf - pao->egul) / 0xffffffff;
    /* copied up to here */

      
    if( iInitRead == INIT_FROM_HARDWARE  &&  pao->dpvt != NULL){
      /* set linear conversion slope */
       pai.eslo = pao->eslo;
       /* pvAi=pvmeio; */
       pai.inp.value.vmeio.parm   = pvmeio->parm;
       pai.inp.value.vmeio.card   = pvmeio->card;
       pai.inp.value.vmeio.signal = pvmeio->signal;
       pai.dpvt   = pao->dpvt;

       sprintf(pai.name,"%s",pao->name);
       
       read_ai(&pai);

       pao->rval=pai.rval & ((struct recDesc *)(pai.dpvt))->uiMask; /* is alreday masked but just in case */ 
        if(pao->linr == 1) 
	  pao->val = pao->eslo * pao->rval + pao->egul;
	else
          pao->val=pao->rval;
    }

  pao->udf = FALSE; /* Bob said that :reason already known */
  pao->stat=  NO_ALARM;
  pao->sevr=  NO_ALARM;
  /* Not possible to read last output value ! */

  if(iInitRead == INIT_RESTORE_VAL)
    return(2); /* no conversion */
  else
    return(Ok);
}

/*---------------------------*/

STATIC long write_ao(struct  aoRecord *pao)
{
  struct vmeio *pvmeio = &pao->out.value.vmeio;
  int status=0;
  unsigned int iVal=0;

  
 if(pao->linr == 1) 
   pao->rval = (long)((pao->val - pao->egul) / pao->eslo); 
 else
   pao->rval = (long) pao->val;

 iVal=pao->rval & ((struct recDesc *)(pao->dpvt))->uiMask;
   status = WRITE_AO_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pao->dpvt,REC_AO,pao->name, iVal);
  
  if (status != Ok) {
	recGblSetSevr(pao, WRITE_ALARM, INVALID_ALARM);
  }

  return(Ok); 
}

/*---------------------------*/

STATIC long linconv_ao(
	struct aoRecord *pao,
	int after
)
{
 
  if (!after)
	return(Ok);  

  pao->eslo = (pao->eguf - pao->egul) /(float) 0xFFFFFFFF;

  return(Ok);
}
/**********************/
/*        BI         */
/*********************/

STATIC long get_ioint_bi_info(int cmd,struct biRecord *pbi, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      

  if( pbi->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pbi->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pbi->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }
  return (0);

}

/*------------------------------------*/

STATIC long init_bi_record(struct biRecord *pbi)
{
  struct recDesc  *pRecDesc;
  int iInitRead=0;
  struct vmeio *pvmeio = &pbi->inp.value.vmeio;
  short sValTmp=0;

  if (pbi->inp.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pbi,
                "devbiGPMM (init_record) Illegal INP field");
        return(S_db_badField);
  }
 
  PRINT_DEBUG("REC BI  %s",pbi->name);
  if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("BI %s Memory C=%d S=%d allocation error\n",pbi->name, pvmeio->card, pvmeio->signal);
	pbi->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("BI record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */
        getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	sValTmp=pRecDesc->cShiftBits;
	pRecDesc->cShiftBits= PARAM_VLIDITY (sValTmp,0,16,0); 
	
	pbi->dpvt=(char *)pRecDesc; 
	PRINT_DEBUG("REC BI %s: shft=%d\n",pbi->name,(short) pRecDesc->cShiftBits);
      }
            
    }
  

  pbi->udf = FALSE; /* Bob said that :reason already known */

  return(Ok);
}

/*--------*/

STATIC long read_bi(struct biRecord *pbi)
{
  struct vmeio *pvmeio = &pbi->inp.value.vmeio;
  unsigned short val;
  int status;
  
  status = READ_BI_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pbi->dpvt, pbi->name, &val);

  	if (status != Ok) {
	
	  recGblSetSevr(pbi, READ_ALARM, INVALID_ALARM);
	}
	else{
	  pbi->rval = val;
	  pbi->val  = pbi->rval;
	}
	
  return(Ok);
}

/**********************/
/*        BO         */
/*********************/

STATIC long get_ioint_bo_info(int cmd,struct boRecord *pbo, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      


  if( pbo->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pbo->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pbo->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }

  return (0);

}

/*------------------------------------*/

STATIC long init_bo_record(struct boRecord *pbo)
{
  
  struct biRecord bi1;
  struct recDesc  *pRecDesc=NULL;
  int iInitRead=0;
  struct vmeio *pvmeio = &pbo->out.value.vmeio;
  short sValTmp=0;

  if (pbo->out.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pbo,
                "devBoGPMM (init_record) Illegal OUT field");
        return(S_db_badField);
  }
  PRINT_DEBUG("REC BO %s: \n",pbo->name);

if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("BO DIR record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pbo->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;
	pRecDesc->pRegLocation=NULL;

	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("BO record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */
	getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
        sValTmp=pRecDesc->cShiftBits;
	pRecDesc->cShiftBits= PARAM_VLIDITY (sValTmp,0,16,0);
	 
	pbo->dpvt=(char *)pRecDesc; 
      }
      
    }

 if( iInitRead == INIT_FROM_HARDWARE && pbo->dpvt != NULL){
       bi1.inp.value.vmeio.parm   = pvmeio->parm;
       bi1.inp.value.vmeio.card   = pvmeio->card;
       bi1.inp.value.vmeio.signal = pvmeio->signal; 
       bi1.dpvt   = pbo->dpvt;

       sprintf(bi1.name,"%s",pbo->name);
       
       read_bi(&bi1);

       pbo->rval=bi1.rval & 0x1;
       pbo->val =pbo->rval;
 }

       pbo->udf = FALSE; /* Bob said that :reason already known */
       pbo->stat=  NO_ALARM;
       pbo->sevr=  NO_ALARM;

       if(iInitRead == INIT_RESTORE_VAL)
	 return(2); /* no conversion */
       else
	 return(Ok);
}

/*---------------------------*/

STATIC long write_bo(struct boRecord *pbo)
{
  int status=0;
  struct vmeio *pvmeio = &pbo->out.value.vmeio;

  status=WRITE_BO_GPMM(pvmeio->card, pvmeio->signal,(struct recDesc *) pbo->dpvt, pbo->name, (unsigned short) pbo->rval);

  if (status != Ok) {
	recGblSetSevr(pbo, WRITE_ALARM, INVALID_ALARM);
  }

  return(Ok); /* no conversion */
}



/**********************/
/*   MBBI DIRECT     */
/*********************/

STATIC long get_ioint_mbbiDirect_info(int cmd,struct mbbiDirectRecord *pmbbi, IOSCANPVT *ppvt){

#ifndef linux 
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else
  int iX=0;
#endif

  if( pmbbi->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pmbbi->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */
    iX=((struct recDesc *)pmbbi->dpvt)->iIntrNr;
    if( iX >= 0 )
      if(irqGpmmSup.pioScanPvt[iX])
	 *ppvt = irqGpmmSup.pioScanPvt[iX];
#endif /* linux */
  }
  return (0);

}

/*------------------------------------*/



STATIC long init_mbbiDirect_record(struct mbbiDirectRecord *pmbbi)
{
  struct recDesc  *pRecDesc;
  struct vmeio *pvmeio = &pmbbi->inp.value.vmeio;
  short sValTmp=0;
  int iInitRead=0;

  if (pmbbi->inp.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pmbbi,
                "devMbbiGPMM (init_record) Illegal INP field");
        return(S_db_badField);
  }
 
  PRINT_DEBUG("REC MBBI DIR %s: nobt=%d mask=0x%.8X shft=%d\n",pmbbi->name,pmbbi->nobt,(unsigned int) pmbbi->mask, pmbbi->shft);
  if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("MBBI %s Memory C=%d S=%d allocation error\n",pmbbi->name, pvmeio->card, pvmeio->signal);
	pmbbi->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("MBBI DIR record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */

	if(pmbbi->nobt > 0 || pmbbi->shft > 0){
	  pRecDesc->cMaskBitNr=PARAM_VLIDITY (pmbbi->nobt,1,16,16);
	  pRecDesc->cShiftBits= PARAM_VLIDITY ((short)pmbbi->shft,0,16,0); 
	}
	else
	  {
	    getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	    pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,16,16);
	    sValTmp=pRecDesc->cShiftBits;
	    pRecDesc->cShiftBits=PARAM_VLIDITY (sValTmp,0,16,0);
	  }
	
	pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	pmbbi->shft=0; /* to be sure that there is not conversion in the record support or so */
	pmbbi->dpvt=(char *)pRecDesc; 
	PRINT_DEBUG("REC MBBI %s: mask=%0x.8X nobt=%d shft=%d\n",pmbbi->name,pRecDesc->uiMask,pRecDesc->cMaskBitNr,pRecDesc->cShiftBits);
      }
            
    }
  

  pmbbi->udf = FALSE; /* Bob said that :reason already known */

  return(Ok);
}

/*---------------------------*/

STATIC long read_mbbiDirect(struct mbbiDirectRecord *pmbbi)
{
  struct vmeio *pvmeio = &pmbbi->inp.value.vmeio;
  unsigned short val;
  int status;
  
  status = READ_MBBI_DIRRECT_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pmbbi->dpvt,REC_MBBI_D, pmbbi->name, &val);

  	if (status != Ok) {
	
	  recGblSetSevr(pmbbi, READ_ALARM, INVALID_ALARM);
	}
	else{
	  pmbbi->rval = val & ((struct recDesc *) pmbbi->dpvt)->uiMask;
	  pmbbi->val = pmbbi->rval;
	}
	
  return(2); /* No conversion */
}

/**********************/
/*   MBBO DIRECT     */
/*********************/

STATIC long get_ioint_mbboDirect_info(int cmd,struct mbboDirectRecord *pmbbo, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      


  if( pmbbo->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pmbbo->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pmbbo->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }
  return (0);

}

/*------------------------------------*/


STATIC long init_mbboDirect_record(struct mbboDirectRecord *pmbbo)
{
  
  struct mbbiDirectRecord mbbiDirect1;
  struct recDesc  *pRecDesc=NULL;
  struct vmeio *pvmeio = &pmbbo->out.value.vmeio;
  short sValTmp=0;
  int iInitRead=0;

  if (pmbbo->out.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pmbbo,
                "devMbboGPMM (init_record) Illegal OUT field");
        return(S_db_badField);
  }
  PRINT_DEBUG("REC MBBO DIR %s: nobt=%d mask=0x%.8X shft=%d\n",pmbbo->name,pmbbo->nobt,(unsigned int) pmbbo->mask,(unsigned int) pmbbo->shft);

if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("MBBO DIR record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pmbbo->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;
	pRecDesc->pRegLocation=NULL;

	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("MBBO DIR record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */

	if(pmbbo->nobt > 0 || pmbbo->shft > 0){
	  pRecDesc->cMaskBitNr=PARAM_VLIDITY (pmbbo->nobt,1,16,16);
	  pRecDesc->cShiftBits= PARAM_VLIDITY ((short)pmbbo->shft,0,16,0); 
	}
	else
	  {
	    getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	    pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,16,16);
	    sValTmp=pRecDesc->cShiftBits;
	    pRecDesc->cShiftBits=PARAM_VLIDITY (sValTmp,0,16,0);
	  }
	
	pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	 
	pmbbo->dpvt=(char *)pRecDesc; 
      }
      
    }
 
 if( iInitRead == INIT_FROM_HARDWARE && pmbbo->dpvt != NULL){ 
       mbbiDirect1.inp.value.vmeio.parm   = pvmeio->parm;
       mbbiDirect1.inp.value.vmeio.card   = pvmeio->card;
       mbbiDirect1.inp.value.vmeio.signal = pvmeio->signal; 
       mbbiDirect1.dpvt   = pmbbo->dpvt;
       mbbiDirect1.mask=pmbbo->mask;
       /* mbbiDirect1.shft=pmbbo->shft; */
       mbbiDirect1.nobt=pmbbo->nobt;
       sprintf(mbbiDirect1.name,"%s",pmbbo->name);
       
       read_mbbiDirect(&mbbiDirect1);

       pmbbo->shft=0;
       pmbbo->rval=mbbiDirect1.rval & ((struct recDesc *)(mbbiDirect1.dpvt))->uiMask;
       pmbbo->val=pmbbo->rval;
 }


     pmbbo->udf = FALSE; /* Bob said that :reason already known */
     pmbbo->stat=  NO_ALARM;
     pmbbo->sevr=  NO_ALARM; 

     if(iInitRead == INIT_RESTORE_VAL)
       return(2); /* no conversion */
     else
       return(Ok);
}

/*---------------------------*/

STATIC long write_mbboDirect(struct mbboDirectRecord *pmbbo)
{
  int status=0;
  struct vmeio *pvmeio = &pmbbo->out.value.vmeio;

  status=WRITE_MBBO_DIRRECT_GPMM(pvmeio->card, pvmeio->signal,(struct recDesc *) pmbbo->dpvt, REC_MBBO_D, pmbbo->name, (unsigned short) pmbbo->rval);

  if (status != Ok) {
	recGblSetSevr(pmbbo, WRITE_ALARM, INVALID_ALARM);
  }

  return(Ok); /* no conversion */
}

/*--------------------------------------------------------------------------*/

/**********************/
/*   STRING IN       */
/*********************/

STATIC long get_ioint_strIn_info(int cmd,struct stringinRecord *pStrIn, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      


  if( pStrIn->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pStrIn->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pStrIn->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }

  return (0);

}

/*------------------------------------*/

STATIC long init_strIn_record(struct stringinRecord *pStrIn)
{
  struct recDesc  *pRecDesc;
  int iInitRead=0;
  struct vmeio *pvmeio = &pStrIn->inp.value.vmeio;

  if (pStrIn->inp.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pStrIn,
                "devStrInGPMM (init_record) Illegal INP field");
        return(S_db_badField);
  }
 
  PRINT_DEBUG("REC STR IN%s\n",pStrIn->name);
  if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("STR IN %s Memory C=%d S=%d allocation error\n",pStrIn->name, pvmeio->card, pvmeio->signal);
	pStrIn->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("STR IN record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

        getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	pRecDesc->uiBytesNo= PARAM_VLIDITY ((short)pRecDesc->uiBytesNo,1,39,39); 
	
	
	pStrIn->dpvt=(char *)pRecDesc; 
	PRINT_DEBUG("STR IN %s: Str len=%d \n",pStrIn->name,pRecDesc->uiBytesNo);
      }
            
    }
  

  pStrIn->udf = FALSE; /* Bob said that :reason already known */

  return(Ok);
}

/*---------------------------*/

STATIC long read_strIn(struct stringinRecord *pStrIn)
{
  struct vmeio *pvmeio = &pStrIn->inp.value.vmeio;
  
  int status;
  
  status = READ_STR_IN_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pStrIn->dpvt, pStrIn->name, pStrIn->val);

  	if (status != Ok) {
	
	  recGblSetSevr(pStrIn, READ_ALARM, INVALID_ALARM);
	}
	
	
  return(Ok); 
}

/*--------------------------------*/

/**********************/
/*   WAVE FORM      */
/*********************/

STATIC long get_ioint_wf_info(int cmd,struct waveformRecord *pwf, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	*descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      


  if( pwf->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pwf->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pwf->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }

  return (0);

}

/*------------------------------------*/


static long init_wf_record(pwf)
    struct waveformRecord       *pwf;
{
  struct recDesc  *pRecDesc=NULL;
  int iInitRead=0;
  struct vmeio *pvmeio = &pwf->inp.value.vmeio;

    /* wf.inp must be an VME_IO */
    switch (pwf->inp.type) {
    case (VME_IO) :
        break;
    default :
        recGblRecordError(S_db_badField,(void *)pwf,
                "devWfGPMM (init_record) Illegal INP field");
        return(S_db_badField);
    }

  PRINT_DEBUG("REC WF %s\n",pwf->name);
  if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("WF %s Memory C=%d S=%d allocation error\n",pwf->name, pvmeio->card, pvmeio->signal);
	pwf->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("WF record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);

	if(pwf->ftvl == DBF_FLOAT || pwf->ftvl == DBF_LONG ||  pwf->ftvl == DBF_ULONG)
	  pRecDesc->cMaskBitNr = 4; /* number of bytes per each element of the waveform*/
	else if (pwf->ftvl == DBF_SHORT || pwf->ftvl == DBF_USHORT)
	  pRecDesc->cMaskBitNr = 2; /* number of bytes per each element of the waveform*/
	else if (pwf->ftvl == DBF_CHAR)
	  pRecDesc->cMaskBitNr = 1; /* number of bytes per each element of the waveform*/
	else
	  pRecDesc->cMaskBitNr = 1; /* number of bytes per each element of the waveform*/
        
	if(pwf->nelm > 0 )
	  pRecDesc->uiBytesNo = pwf->nelm *  pRecDesc->cMaskBitNr;
	else
	  pRecDesc->uiBytesNo = 0;
	  
	pwf->dpvt=(char *)pRecDesc; 
	PRINT_DEBUG("WF %s: len=%d \n",pwf->name,pRecDesc->uiBytesNo);
      }
            
    }
  if(pRecDesc != NULL){
    if( iInitRead == INIT_FROM_HARDWARE && pwf->dpvt != NULL && ((pRecDesc->uiParam1 & WF_REC_TYPE) == WF_REC_OUT)  ){ 
       
      ((struct recDesc *) pwf->dpvt)->uiParam1=((struct recDesc *) pwf->dpvt)->uiParam1 & (~((unsigned int)WF_REC_TYPE) | WF_REC_IN);
      read_wf(pwf);
      ((struct recDesc *) pwf->dpvt)->uiParam1=((struct recDesc *) pwf->dpvt)->uiParam1  | WF_REC_OUT;			    
    }
  }

    pwf->udf = FALSE; /* Bob said that :reason already known */
    return(0);
}

/*--------------------------------------*/

static long read_wf(pwf)
    struct waveformRecord       *pwf;
{
  int      status;
  struct vmeio *pvmeio = (struct vmeio *)&(pwf->inp.value);


  if ( (((struct recDesc *) pwf->dpvt)->uiParam1 & WF_REC_TYPE) == WF_REC_OUT){
    if(pwf->nord > pwf->nelm)
      pwf->nord = pwf->nelm;
    ((struct recDesc *)pwf->dpvt)->uiBytesNo = pwf->nord *  ((struct recDesc *)pwf->dpvt)->cMaskBitNr;
    PRINT_DEBUG(" WF record nord = %ld\n",pwf->nord);
    PRINT_DEBUG(" WF record nelm = %ld\n",pwf->nelm);
    status = WRITE_WF_GPMM(pvmeio->card,pvmeio->signal, (struct recDesc *) pwf->dpvt, pwf->name,pwf->bptr );
  }
  else{  /* WF_REC_IN */

    status = READ_WF_GPMM(pvmeio->card,pvmeio->signal, (struct recDesc *) pwf->dpvt, pwf->name,pwf->bptr );
    if(status == 0){
    /* memcpy(pwf->bptr,pfVal,((struct recDesc *) pvmeio->dpvt)->uiBytesNo); */
    pwf->nord = pwf->nelm;/* no_read; */
    }
    else
      pwf->nord =0;
  }  

  return(0);
}

/*------------------------------------------------------------*/
/*   LONG IN  and OUT RECORDS                                           */
/*------------------------------------------------------------*/

/**********************/
/*    LONG IN        */
/*********************/


STATIC long get_ioint_longIn_info(int cmd,struct longinRecord *pLongIn, IOSCANPVT *ppvt){


#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      

  if( pLongIn->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pLongIn->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pLongIn->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }

  return (Ok);

}

/*---------------------------------------*/

STATIC long init_longIn_record(
	struct longinRecord *pLongIn
)
{
  struct vmeio *pvmeio;
  struct recDesc  *pRecDesc;
  int iInitRead=0;
  pvmeio = (struct vmeio *)&(pLongIn->inp.value.vmeio);

 /* ai.inp must be an VME_IO */
  switch (pLongIn->inp.type) {
	case (VME_IO):
		break;
	default:
		recGblRecordError(S_db_badField, (void *) pLongIn,
			"devLongInGPMM (init_record) Illegal INP field");
		return(S_db_badField);
  }
  
   if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("LongIn record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pLongIn->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("LongIn record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */
        getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,32,32);
	pRecDesc->uiBytesNo=PARAM_VLIDITY (pRecDesc->uiBytesNo,1,4,4);
	
	if(pRecDesc->cMaskBitNr>=32)   	
	  pRecDesc->uiMask=0xFFFFFFFF; 
	else                           
	  pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	 
	pLongIn->dpvt=(char *)pRecDesc; 
      }
            
    }
  
  pLongIn->udf = FALSE; /* Bob said that :reason already known */

  return(Ok);
}

/*---------------------------*/

STATIC long read_longIn(struct longinRecord *pLongIn)
{
  struct vmeio *pvmeio;/*  = &pLongIn->inp.value.vmeio; */
  int status=Ok;
  unsigned int iVal=0;

  pvmeio = (struct vmeio *)&(pLongIn->inp.value.vmeio);

  status = READ_AI_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pLongIn->dpvt,REC_LONGIN,pLongIn->name, &iVal);
  if( status == Ok){
    iVal=iVal & ((struct recDesc *)(pLongIn->dpvt))->uiMask;
    pLongIn->val=iVal;
  }
 
	if (status != Ok) {
	
		recGblSetSevr(pLongIn, READ_ALARM, INVALID_ALARM);
	}
	
	return(Ok); 
}

/**********************/
/*      LONG OUT      */
/*********************/

STATIC long get_ioint_info_longOut(int cmd,struct longoutRecord *pLongOut, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      

  if( pLongOut->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pLongOut->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pLongOut->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }
  return (0);

}

/*------------------------------------*/

STATIC long init_record_longOut(struct longoutRecord *pLongOut)
{
  /* int         status=Ok; */
    struct longinRecord pLongIn;
    struct recDesc  *pRecDesc=NULL;
    int iInitRead=0;
    struct vmeio *pvmeio = &pLongOut->out.value.vmeio;

 /* ao.out must be an VME_IO */
    if (pLongOut->out.type != VME_IO) {
  
		recGblRecordError(S_db_badField, (void *) pLongOut,
			"devLongOutGPMM (init_record) Illegal OUT field");
		return(S_db_badField);
  }

    /* copy region begins here */
    if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("Long Out record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pLongOut->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("Long Out record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */
        getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,32,32);
	pRecDesc->uiBytesNo=PARAM_VLIDITY (pRecDesc->uiBytesNo,1,4,4);

	if(pRecDesc->cMaskBitNr>=32)   
	  pRecDesc->uiMask=0xFFFFFFFF; 
	else                           	
	  pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	 
	pLongOut->dpvt=(char *)pRecDesc; 
      }
      
    }

    /* copied up to here */

      
    if( iInitRead == INIT_FROM_HARDWARE  &&  pLongOut->dpvt != NULL){
      
       /* pLongIn=pvmeio; */
       pLongIn.inp.value.vmeio.parm   = pvmeio->parm;
       pLongIn.inp.value.vmeio.card   = pvmeio->card;
       pLongIn.inp.value.vmeio.signal = pvmeio->signal;
       pLongIn.dpvt   = pLongOut->dpvt;

       sprintf(pLongIn.name,"%s",pLongOut->name);
       
       read_longIn(&pLongIn);

       pLongOut->val=pLongIn.val & ((struct recDesc *)(pLongIn.dpvt))->uiMask; /* is alreday masked but just in case */ 
        
    }

  pLongOut->udf = FALSE; /* Bob said that :reason already known */
  pLongOut->stat=  NO_ALARM;
  pLongOut->sevr=  NO_ALARM;
  /* Not possible to read last output value ! */

  if(iInitRead == INIT_RESTORE_VAL)
    return(2); /* no conversion */
  else
    return(Ok);
}

/*---------------------------*/

STATIC long write_longOut(struct  longoutRecord *pLongOut)
{
  struct vmeio *pvmeio = &pLongOut->out.value.vmeio;
  int status=0;
  unsigned int iVal=0;

 iVal=pLongOut->val & ((struct recDesc *)(pLongOut->dpvt))->uiMask;
   status = WRITE_AO_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pLongOut->dpvt,REC_LONGOUT,pLongOut->name, iVal);
  
  if (status != Ok) {
	recGblSetSevr(pLongOut, WRITE_ALARM, INVALID_ALARM);
  }

  return(Ok); 
}

/*---------------------------*/

/**********************/
/*   MBBI      */
/*********************/

STATIC long get_ioint_mbbi_info(int cmd,struct mbbiRecord *pmbbi, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      


  if( pmbbi->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pmbbi->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pmbbi->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }
  return (0);

}

/*------------------------------------*/



STATIC long init_mbbi_record(struct mbbiRecord *pmbbi)
{
  struct recDesc  *pRecDesc;
  struct vmeio *pvmeio = &pmbbi->inp.value.vmeio;
  short sValTmp=0;
  int iInitRead=0;

  if (pmbbi->inp.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pmbbi,
                "devMbbiGPMM (init_record) Illegal INP field");
        return(S_db_badField);
  }
 
  PRINT_DEBUG("REC MBBI DIR %s: nobt=%d mask=0x%.8X shft=%d\n",pmbbi->name,pmbbi->nobt,(unsigned int) pmbbi->mask, pmbbi->shft);
  if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("MBBI %s Memory C=%d S=%d allocation error\n",pmbbi->name, pvmeio->card, pvmeio->signal);
	pmbbi->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;

	pRecDesc->pRegLocation=NULL;
	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("MBBI DIR record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */

	if(pmbbi->nobt > 0 || pmbbi->shft > 0){
	  pRecDesc->cMaskBitNr=PARAM_VLIDITY (pmbbi->nobt,1,16,16);
	  pRecDesc->cShiftBits= PARAM_VLIDITY ((short)pmbbi->shft,0,16,0); 
	}
	else
	  {
	    getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	    pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,16,16);
	    sValTmp=pRecDesc->cShiftBits;
	    pRecDesc->cShiftBits=PARAM_VLIDITY (sValTmp,0,16,0);
	  }
	
	pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	pmbbi->shft=0; /* to be sure that there is not conversion in the record support or so */
	pmbbi->dpvt=(char *)pRecDesc; 
	PRINT_DEBUG("REC MBBI %s: mask=%0x.8X nobt=%d shft=%d\n",pmbbi->name,pRecDesc->uiMask,pRecDesc->cMaskBitNr,pRecDesc->cShiftBits);
      }
            
    }
  

  pmbbi->udf = FALSE; /* Bob said that :reason already known */

  return(Ok);
}

/*---------------------------*/

STATIC long read_mbbi(struct mbbiRecord *pmbbi)
{
  struct vmeio *pvmeio = &pmbbi->inp.value.vmeio;
  unsigned short val;
  int status;
  
  status = READ_MBBI_DIRRECT_GPMM(pvmeio->card, pvmeio->signal, (struct recDesc *) pmbbi->dpvt, REC_MBBI, pmbbi->name, &val);

  	if (status != Ok) {
	
	  recGblSetSevr(pmbbi, READ_ALARM, INVALID_ALARM);
	}
	else{
	  pmbbi->rval = val & ((struct recDesc *) pmbbi->dpvt)->uiMask;
	  pmbbi->val = pmbbi->rval;
	}
	
  return(2); /* No conversion */
}

/**********************/
/*   MBBO      */
/*********************/

STATIC long get_ioint_mbbo_info(int cmd,struct mbboRecord *pmbbo, IOSCANPVT *ppvt){

#ifndef linux
  struct io_GPMM	  *descrp=NULL;
  struct register_Node  *regNode=NULL;
#else       
  int iX=0; 
#endif      

  if( pmbbo->dpvt != NULL){
#ifndef linux
    regNode=((struct recDesc *)pmbbo->dpvt)->pRegLocation;
    descrp=regNode->pCard;
    *ppvt = descrp->paioscanpvt;
#else /* linux defined */                       
    iX=((struct recDesc *)pmbbo->dpvt)->iIntrNr;  
    if( iX >= 0 )                               
      if(irqGpmmSup.pioScanPvt[iX])             
         *ppvt = irqGpmmSup.pioScanPvt[iX];     
#endif /* linux                               */
  }
  return (0);

}

/*------------------------------------*/


STATIC long init_mbbo_record(struct mbboRecord *pmbbo)
{
  
  struct mbbiRecord mbbi1;
  struct recDesc  *pRecDesc=NULL;
  struct vmeio *pvmeio = &pmbbo->out.value.vmeio;
  short sValTmp=0;
  int iInitRead=0;

  if (pmbbo->out.type != VME_IO) {
	recGblRecordError(S_db_badField, (void *) pmbbo,
                "devMbboGPMM (init_record) Illegal OUT field");
        return(S_db_badField);
  }
  PRINT_DEBUG("REC MBBO DIR %s: nobt=%d mask=0x%.8X shft=%d\n",pmbbo->name,pmbbo->nobt,(unsigned int) pmbbo->mask,(unsigned int) pmbbo->shft);

if (pvmeio->parm!=NULL)
    {
      pRecDesc = calloc (1, sizeof(struct recDesc) );

      if (  pRecDesc == NULL ) {
	printf ("MBBO DIR record Memory C=%d S=%d allocation error\n", pvmeio->card, pvmeio->signal);
	pmbbo->dpvt=NULL;
	return(S_db_badField);
      }
      else{
	
	pRecDesc->sDesc=pvmeio->parm;
	pRecDesc->pRegLocation=NULL;

	if( getRegLocation(pvmeio->card,pvmeio->parm,&(pRecDesc->pRegLocation)) != Ok )
	  printf("MBBO DIR record has no association with the register memory C=%d Parm %s\n", pvmeio->card,pvmeio->parm);

	/* get the poniter here to the Register memory location */

	if(pmbbo->nobt > 0 || pmbbo->shft > 0){
	  pRecDesc->cMaskBitNr=PARAM_VLIDITY (pmbbo->nobt,1,16,16);
	  pRecDesc->cShiftBits= PARAM_VLIDITY ((short)pmbbo->shft,0,16,0); 
	}
	else
	  {
	    getRecordParam(pvmeio->parm,pRecDesc,&iInitRead);
	    pRecDesc->cMaskBitNr=PARAM_VLIDITY (pRecDesc->cMaskBitNr,1,16,16);
	    sValTmp=pRecDesc->cShiftBits;
	    pRecDesc->cShiftBits=PARAM_VLIDITY (sValTmp,0,16,0);
	  }
	
	pRecDesc->uiMask=(((unsigned long)1<<pRecDesc->cMaskBitNr)-1);      /* calculate the mask out of the nr ob bits */
	 
	pmbbo->dpvt=(char *)pRecDesc; 
      }
      
    }
 
 if( iInitRead == INIT_FROM_HARDWARE && pmbbo->dpvt != NULL){ 
       mbbi1.inp.value.vmeio.parm   = pvmeio->parm;
       mbbi1.inp.value.vmeio.card   = pvmeio->card;
       mbbi1.inp.value.vmeio.signal = pvmeio->signal; 
       mbbi1.dpvt   = pmbbo->dpvt;
       mbbi1.mask=pmbbo->mask;
       /* mbbi1.shft=pmbbo->shft; */
       mbbi1.nobt=pmbbo->nobt;
       sprintf(mbbi1.name,"%s",pmbbo->name);
       
       read_mbbi(&mbbi1);

       pmbbo->shft=0;
       pmbbo->rval=mbbi1.rval & ((struct recDesc *)(mbbi1.dpvt))->uiMask;
       pmbbo->val=pmbbo->rval;
 }


     pmbbo->udf = FALSE; /* Bob said that :reason already known */
     pmbbo->stat=  NO_ALARM;
     pmbbo->sevr=  NO_ALARM; 

     if(iInitRead == INIT_RESTORE_VAL)
       return(2); /* no conversion */
     else
       return(Ok);
}

/*---------------------------*/

STATIC long write_mbbo(struct mbboRecord *pmbbo)
{
  int status=0;
  struct vmeio *pvmeio = &pmbbo->out.value.vmeio;

  status=WRITE_MBBO_DIRRECT_GPMM(pvmeio->card, pvmeio->signal,(struct recDesc *) pmbbo->dpvt, REC_MBBO, pmbbo->name, (unsigned short) pmbbo->rval);

  if (status != Ok) {
	recGblSetSevr(pmbbo, WRITE_ALARM, INVALID_ALARM);
  }

  return(Ok); /* no conversion */
}

/*--------------------------------------------------------------------------*/
