/* this is the drvier support for the memory mapped VME cards and Linux  */
/* The driver code was writte by M.Dach 2005 - 2008 */
#ifndef linux
#include <vxWorks.h>
#include <vme.h>
#include <rebootLib.h>
#include <vxLib.h>
#include <taskLib.h>
#include <intLib.h>
#include <stdlib.h>
#include <sysLib.h>
#include <logLib.h>
#include <iv.h>

#else /* linux */
#include "epicsExport.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <pthread.h>
#endif

#include <stdio.h>

/* #include <dbScan.h> */
#include <aiRecord.h>
#include <aoRecord.h>
#include <drvSup.h>
#include <devLib.h>
#include "slsDevLib.h"
#include <string.h>
#include <epicsVersion.h>

#include <drvGPMM.h>
#include <userGPMM.h>

#ifdef linux
#include "xgpmm.h"
extern struct sIrqGpmmSup irqGpmmSup;
static pthread_t        gpmmIrqThread;
/*-------------------------------------------------------*/
/*    IRQ routine for user space                         */
/*-------------------------------------------------------*/
/* This routine requires the kernel Linux module which
   acts as a ISR (Interrupt Service Routine). 
   The gpmmIrqHandlerThread routine uses the blocking read 
   on a file descriptor. This read command gels one integer
   number (iTemp varaible). Every bit of that variable
   refers to the given interrupt number (ex. 0x0009 
   mens that two interrupts have occured: 1 and 8 ).
*/
static void *gpmmIrqHandlerThread (void *ptr){
  
  unsigned int iTemp=0;
  int iLen=0;
  int iX=0;

  iLen=sizeof(unsigned int)*8;
  /* printf("Entering the endless loop of the gpmmIrqHandlerThread\n"); */
  while(1){
    read(irqGpmmSup.gpmmDrvFD, &iTemp, sizeof(unsigned int));
    /* printf("FD %d IRQ 0x%.8x received \n",irqGpmmSup.gpmmDrvFD,iTemp); */
    if(iTemp>0){
      for(iX=0;iX<iLen;iX++){
	if(iTemp & 0x0001){
	  /* printf("IRQ nr -> %d has occured \n",iX); */
	  if(irqGpmmSup.pioScanPvt[iX]){
	    scanIoRequest(irqGpmmSup.pioScanPvt[iX]);
	    irqGpmmSup.piIrqCount[iX] ++; 
	       } 
	}
	iTemp=iTemp>>1;
      }/* for loop */
    }/* got valid data from the driver */

  }/* endless loop */
  return(NULL);
}
#endif

#define PARAMS_NOT_VALID(LIST,NODE,CARD) \
(NODE = slsDevPrivate(slsDevNodeByCardNum(LIST,CARD))) == NULL ? 1 \
: (NODE->configured == FALSE) ? 1 : 0


static long DRIVER_INIT();
static long IO_REPORT(); 

/*
 *	Driver entry table
 */
struct {
    long number;
    long (*report)();
    long (*init)();
} drvGPMM = {
    2,
    IO_REPORT,
    DRIVER_INIT
};

#ifdef linux
epicsExportAddress(drvet,drvGPMM);
#endif

/* 
 *	pointers to the binary output cards 
 */
STATIC struct devInfo *pGPMMList = NULL;


STATIC char *id_expected = "GPMM board";

/*
 *  Lists the board configuration info
 */
void GPMMList()
{
    slsDevList( id_expected, pGPMMList );
}

/*---------------------------------------------------------*/
/* getGPMMDevBlock(int cardNumber)                         */
/* For the given card number, it returns a pointer to      */
/* the struct devInfo                                      */
/* that structure conatins all of the information about    */  
/* given card and asociated registers                      */
/* It returns NULL if the given CARD number does           */
/* not correspond to any DevBlock                          */
/*---------------------------------------------------------*/

struct devInfo *getGPMMDevBlock(int iCardNumber) {
  
  struct devInfo        *pDevBlock=NULL;
  
  pDevBlock = slsDevFirst( pGPMMList );
  while (pDevBlock != NULL) {
    if(slsDevCardNum(pDevBlock)==iCardNumber)
      break;
    pDevBlock=slsDevNext(pDevBlock);
  }

  return pDevBlock;
}


/*---------------------------------------------------------*/
/* defaultGPMMISR is a default Inerrupt Service Routine    */
/*                                                         */
/* It does the basic things in order to                    */ 
/* trigger EPICS channels when interrupt occurs            */
/*---------------------------------------------------------*/

void defaultGPMMISR(struct io_GPMM  *pPrivat){

  pPrivat->iIntrCounter++;
  scanIoRequest(pPrivat->paioscanpvt);
}


/*---------------------------------------------------------*/
/* addGPMMInterrupt registers the interrupts for EPICS    */
/* channels                                                */
/*---------------------------------------------------------*/
#ifndef linux
void addGPMMInterrupt( int iCardNr, int iIntLevel, int iVector, void (*gpmmUserISR)(struct io_GPMM *pPrivate)){
  
  struct devInfo        *pDevBlock;
  struct io_GPMM	  *pPrivate;

  pDevBlock = slsDevFirst( pGPMMList );
  while (pDevBlock != NULL) {
    if(slsDevCardNum(pDevBlock)==iCardNr)
      break;
    pDevBlock=slsDevNext(pDevBlock);
  }

  if(pDevBlock != NULL){
    pPrivate = slsDevPrivate(pDevBlock);
    scanIoInit(&(pPrivate->paioscanpvt));
    /* printf("paioscanpvt = %p Registered\n",pPrivate->paioscanpvt); */
    if(intConnect (INUM_TO_IVEC(iVector),gpmmUserISR,(int)pPrivate)!= OK){
      printf("Interrupts not register corectly for card nr %d\n",iCardNr);
      
    }
    else{
      if(sysIntEnable(iIntLevel)!= OK){
	printf("Interrupt level  not accepted  for card nr %d\n",iCardNr);
      }
      else
	printf("ISR for the board %s registered sucessfully\n",pPrivate->pcBoardName);
      pDevBlock->itrLevel=iIntLevel;
      pDevBlock->vectorNum=iVector;
    }
 
  }
  else{
    printf("Wrong card number for Interrupt registartion\n");
  }
}
#else /* linux defined */
void addGPMMInterrupt( int iCardNr, int iIntrNr, unsigned int uiBaseAddr, void (*gpmmUserISR)(struct io_GPMM *pPrivate)){

  struct stData_IRQ ssData_IRQ;

    if(!irqGpmmSup.irqHandlerOn){
        irqGpmmSup.gpmmDrvFD = open("/dev/xgpmm", O_RDONLY);
 
	if( irqGpmmSup.gpmmDrvFD == -1) {
	  printf("/dev/xgpmm open error...\n");
	  return;
	} 
       
      irqGpmmSup.irqHandlerOn=1;
    } /* if !irqGpmmSup.irqHandlerOn */
    
    if(irqGpmmSup.pioScanPvt[iIntrNr]==NULL){
      scanIoInit(&(irqGpmmSup.pioScanPvt[iIntrNr]));
      /* register the irq handler in linux kernel space */
      ssData_IRQ.iIrqNumber=iIntrNr;
      ssData_IRQ.iDevBaseAddress=uiBaseAddr;
      ioctl(irqGpmmSup.gpmmDrvFD,XGPMM_IOCTL_IRQ_SET,&ssData_IRQ);
    }
    else{
      printf("IRQ %d already registered registered\n",iIntrNr);
      return;
    }


}
#endif /* linux */

/*---------------------------------------------------------*/
/* GPMMConfigure registers the card in the VME memory      */
/*                                                         */
/* caches board info for use during driver init            */
/* channels                                                */
/*---------------------------------------------------------*/


void GPMMConfigure(
    int         cardNum,          /* card number as used in INP fields */
    void        *base,            /* base address of the board */
    char        *pcBoardName,      /* name of the board */
    char        *sAddrAccessMode, /* A16 A24 A32 */
    char        *sDataAccessMode  /* D8 D16 D32 */			
)
{
    int         vectorNum=0;      /* interrupt vector number. If 0 then one */
				
    int         itrLevel=0;       /* board's interrupt level */
    /* int         itrState=0;        whether or not interrupts should be */
    struct io_GPMM	*pPrivate;

    if ( slsDevInUse( base, cardNum, &pGPMMList ) ) {
#ifndef linux
	logMsg("%s: card %2d or address %x already used!\n",
		(int)id_expected,(int)cardNum, (int)base, 0,0,0);
#else
	fprintf(stderr,"devGPMM: card %2d or address %p already used!\n",cardNum,base);
#endif
	return;
    }
    

/*
 *  Allocate space for the device descriptors
 */
    pPrivate = calloc (1, sizeof(struct io_GPMM) );

    if ( pPrivate == NULL ) {
#ifndef linux
	logMsg ("%s: Memory allocation error\n", (int)id_expected,0,0,0,0,0 );
#else
	fprintf(stderr,"devGPMM: Memory allocation error\n");
#endif
	return;
    }

    pPrivate->pRegisterNode     = NULL;
    sprintf(pPrivate->pcBoardName,"%.15s",pcBoardName);
    if(!strcmp(sAddrAccessMode,"A16"))
       pPrivate->sAddrMode=atVMEA16;
    else if(!strcmp(sAddrAccessMode,"A24"))
       pPrivate->sAddrMode=atVMEA24;
    else if(!strcmp(sAddrAccessMode,"A32"))
       pPrivate->sAddrMode=atVMEA32;
    else{
	 printf("Address access mode not defined correctly.\nThe correct one is: A16 or A24 or A32\n");
	 return;
       }

    if(!strcmp(sDataAccessMode,"D8"))
       pPrivate->sDataMode=VMED8;
    else if(!strcmp(sDataAccessMode,"D16"))
       pPrivate->sDataMode=VMED16;
    else if(!strcmp(sDataAccessMode,"D32"))
       pPrivate->sDataMode=VMED32;
    else{
	 printf("data access mode not defined correctly.\nThe correct one is: D8 or D16 or D32\n");
	 return;
       }
    
    pPrivate->paioscanpvt     = NULL;
    pPrivate->iIntrCounter    = 0;
    pPrivate->configured      = FALSE; 
    /* pPrivate->enableItrOnInit = itrState; */

    slsDevConfigure( id_expected, cardNum, (void *)base, vectorNum, itrLevel, 
		    pPrivate, &pGPMMList );
}



/*---------------------------------------------------------*/
/* addGPMMRegister defines the register  to be used be     */
/* EPICS channels                                          */
/* register is a memory loaction with the respect to       */
/* the base addres defined in  GPMMConfigure function      */
/*---------------------------------------------------------*/

int addGPMMRegister(
	     unsigned int         cardNum,        /* card number as used in INP fields */
	     unsigned short       iRefNr,         /* signal number as used in INP field */
	     unsigned int         iAddrOffset,    /* ofset from the base addres */
	     unsigned int       uiBytesNr,       /* size of the register in Bytes */
	     char                 cRW,            /* data to be read (R) written (W) or nothing (X) */
	     char                *sDesc,          /* register description 20 char's */
             int                 (*fnct)(unsigned short, unsigned short, struct recDesc *, unsigned short, unsigned short, char *, void *)
	     )
{
    struct io_GPMM	 *descrp;
    struct register_Node *regNode;

    if ( PARAMS_NOT_VALID(pGPMMList,descrp,cardNum) ) {
      if (descrp == NULL) return S_drvGPMM_noDevice;
    }
    
   
    if ( descrp->pRegisterNode != NULL) {
      /* printf("NODE %s\n",descrp->pRegisterNode->sDesc); */
	regNode=descrp->pRegisterNode;

    while ( regNode->nextNode != NULL){
      /* printf("node %s\n",regNode->sDesc); */
        regNode=regNode->nextNode;
    }  
       regNode->nextNode = calloc (1, sizeof(struct register_Node) );
     if ( regNode->nextNode != NULL )
       regNode=regNode->nextNode;
    }

    else{
      descrp->pRegisterNode = calloc (1, sizeof(struct register_Node) );
      regNode=descrp->pRegisterNode;
    }
          
    if ( regNode != NULL ){
        regNode->iRefNr = iRefNr;
        regNode->iAddrOffset = iAddrOffset;
	sprintf(regNode->sDesc,"%.50s",sDesc);
        regNode->pLocalAddr = NULL;
	regNode->uiBytesNr   = uiBytesNr;
	if(uiBytesNr<=0) printf("Error 4th param i.e. nr of bytes should be positive!\n");
	regNode->cRW        = cRW;
	regNode->fnct       = fnct;
	regNode->pCard      = NULL; /* so far there is no association done with the card */
	                            /* this assoiation is done in the init function */
        regNode->sConfigured= FALSE;
        regNode->nextNode=NULL;
      }

    return Ok;
    
}

#ifdef linux

void *pmmapio(off_t phy_addr, size_t phy_length)
{
  void *pmem;
  int   fd;
                                                                                
  /*  if (phy_addr % MAP_PAGESIZE)
    {
        fprintf(stderr, "physical address error!\n");
        return NULL;
    }
    if (phy_length % MAP_PAGESIZE)
    {
        fprintf(stderr, "physical length error!\n");
        return NULL;
    }
  */
    if ((fd = open("/dev/mem", O_RDWR | O_SYNC)) < 0)
    {
        fprintf(stderr, "opening /dev/mem failed!\n");
        return NULL;
    }
                                                                                
                                                                                
    pmem = mmap(NULL,                   /* start           */
            phy_length,                 /* length          */
            (PROT_READ | PROT_WRITE),   /* protection      */
            MAP_SHARED,                 /* flags           */
            fd,                         /* file descriptor */
            phy_addr);                  /* offset  phy_addr        */
                                                                                
                                                                                
                                                                                
    if ((pmem == MAP_FAILED) || (pmem == NULL))
    {
        fprintf(stderr, "mmaping /dev/mem failed!\n");
        return NULL;
    }
                                                                                
    if (close(fd))
    {
        fprintf(stderr, "closing /dev/mem failed!\n");
    }
     
    return pmem;
}
#endif
/*-------------------------------------------------------------------------------------*/

static long DRIVER_INIT()
{
  struct devInfo        *pCard;
  struct io_GPMM         *pPrivate;
  struct register_Node  *regNode;

  int                    status=0;
#ifndef linux
    int                    iX=0;  /* index for for loops */
    char                  pcProbe[8];
    short                 sBytesNrProbe=0; 
    short                 sMemProbFailFlag=0;
    short                 sMemProbFailCnt=0;
    int                   iMemProbMode=0;
#else
    size_t         ulMapSize=0;
    size_t         ulMapNew=0;
    char         *pLocalAddrMain=NULL;
#endif

  short                 sProbe=0;
    

  pCard = slsDevFirst(pGPMMList);
    
  while (pCard != NULL) {

    pPrivate = slsDevPrivate( pCard );

#ifdef linux
    /* for linux we have to deterime the size of the area which will be mapped to the loacal one */
    /* so the maping process is done in two steps */
    regNode=pPrivate->pRegisterNode;
    ulMapSize=0;
    while( regNode != NULL){
  
      ulMapNew=regNode->iAddrOffset+regNode->uiBytesNr*sizeof(char);
      if(ulMapNew>ulMapSize)
	ulMapSize=ulMapNew;

      regNode = regNode->nextNode;
    }
    
    pLocalAddrMain=pmmapio(slsDevAddrs(pCard),ulMapSize);
	  
	  if(pLocalAddrMain==NULL){
	    printf( "%s: Cannot register device %s at 0x%x \n",
		    id_expected,  pPrivate->pcBoardName, slsDevAddrs(pCard) );
	    
	  }
   PRINT_DEBUG("Board Name %s Local address %p Memory Area size %d\n",pPrivate->pcBoardName,pLocalAddrMain,ulMapSize);

#endif
    /*-----------  register the card in the A32 address space -------------------*/
    regNode=pPrivate->pRegisterNode;
    while( regNode != NULL){
      regNode->pCard=(void *) pPrivate;
      regNode->sConfigured=TRUE;
#ifndef linux
#if (defined(EPICS_REVISION) && (EPICS_REVISION == 13) && (EPICS_MODIFICATION <= 10))

      if ( (status = devRegisterAddress((const char *)pPrivate->pcBoardName , pPrivate->sAddrMode,
					(void *)&((unsigned char *) slsDevAddrs(pCard))[regNode->iAddrOffset],
					regNode->uiBytesNr*sizeof(char),
					(void **) &regNode->pLocalAddr)
	    != OK )) {
#else
	if ( (status = devRegisterAddress((const char *)pPrivate->pcBoardName , pPrivate->sAddrMode,
					  (size_t)&((unsigned char *) slsDevAddrs(pCard))[regNode->iAddrOffset],
					  regNode->uiBytesNr*sizeof(char),
					  (volatile void **) &regNode->pLocalAddr)
	      != OK )) {

#endif /* EPICS_REVISION */
#else  /* linux */
	
	  /* regNode->pLocalAddr=pmmapio(slsDevAddrs(pCard) + regNode->iAddrOffset,regNode->uiBytesNr*sizeof(char));
	   printf("%s Base Address 0X%x Local one %p = 0x%x\n",regNode->sDesc,slsDevAddrs(pCard) + regNode->iAddrOffset,regNode->pLocalAddr,*((int *) regNode->pLocalAddr)); */
	  if(pLocalAddrMain!=NULL){
	    regNode->pLocalAddr=&(pLocalAddrMain[regNode->iAddrOffset]);
	    PRINT_DEBUG("Register %s Local address %p Memory Area size %d\n",regNode->sDesc,regNode->pLocalAddr,regNode->uiBytesNr*sizeof(char));
	  }
	  else{
	    /* if(regNode->pLocalAddr==NULL){ */
		     
#endif /* linux */
	     
	    printf( "%s: Cannot register %s at 0x%x. Error is %x\n",
		    id_expected,  regNode->sDesc, (unsigned int)slsDevAddrs(pCard) + regNode->iAddrOffset, status );
	    
	    regNode->sConfigured=FALSE;
	  }

	  /*--------  Verify that something exists at the specified address -----------*/
 
	   
	  sProbe=regNode->uiBytesNr*sizeof(char);
	  /* if(sProbe>20) sProbe=20;  this depends on the size of the pcProbe array */

#ifndef linux	   
	   if(sProbe>0){
	     if(pPrivate->sDataMode == VMED8)
	       sBytesNrProbe=1;
             else if(pPrivate->sDataMode == VMED16)
               sBytesNrProbe=2;
             else if(pPrivate->sDataMode == VMED32)
               sBytesNrProbe=4;
             else
	       sBytesNrProbe=1; /* Just in case */

	     sProbe=sProbe/sBytesNrProbe;
	     sMemProbFailFlag=0;
	     sMemProbFailCnt=0;
	     if( regNode->cRW == 'R')
	       iMemProbMode=VX_READ;
             else if(regNode->cRW == 'W')
	       iMemProbMode=VX_WRITE;

	       
	     if(regNode->cRW == 'R' || regNode->cRW == 'W'){
	      for(iX=0;iX<sProbe;iX++){
		
	       if((status = vxMemProbe(&(((char *)regNode->pLocalAddr)[iX*sBytesNrProbe]),iMemProbMode,sBytesNrProbe,(char *)pcProbe)) != OK)
		 {
		   if(regNode->sConfigured==TRUE)
		     printf("%s: Register %s : vxMemProbe failed to %s at 0x%x :",pPrivate->pcBoardName,regNode->sDesc,regNode->cRW=='R'? "READ":"WRITE",(unsigned int)slsDevAddrs(pCard) + regNode->iAddrOffset );
		   if(sMemProbFailFlag==0)
		     printf(" [0x%x",( iX * sBytesNrProbe ));
		   sMemProbFailCnt++;
		   sMemProbFailFlag=1;
		   regNode->sConfigured=FALSE;
		   
		 }
	       else{
		 if(sMemProbFailFlag==1){
		   if(sMemProbFailCnt>1)
		     printf("-0x%x]",( (iX-1) * sBytesNrProbe ));
		   else
		     printf("]");
		 }
		 sMemProbFailFlag=0;
		 sMemProbFailCnt=0;
	       }
	      }/* for loop */

	      if(sMemProbFailFlag==1){
		if(sMemProbFailCnt>1)
		  printf("-0x%x]",( (iX-1) * sBytesNrProbe ));
		else
		  printf("]");
	      }
	      if(regNode->sConfigured==FALSE)
		printf("\n");

	     } /* end of regNode->cRW == 'R' or 'W'*/
 	     
	   } /* if sProbe > 0 */
#endif /* linux */

	  regNode = regNode->nextNode;
	}

	
	/* THIS must be here to go to the next card  in the linked list */
	pPrivate->configured = TRUE; /* this line only sais that the given */
	                             /* card was configured correctly */
	pCard =  slsDevNext(pCard);  /* take new card structure to configure */
      }
#ifdef linux
      /* launche the thread which deals with Interupts */
      if(irqGpmmSup.irqHandlerOn){
	pthread_create(&gpmmIrqThread, NULL,gpmmIrqHandlerThread,(void *)NULL);
	/* pthread_detach(gpmmIrqThread); it was just to try if possible to terminate the bloced thread */
	printf("Launching the IRQ handler \n");
      }
#endif
    
      return(Ok);
    }
  
  

/*------------------------------------------------------*/
/*               Analog Input write                     */
/* This function deals with the mask fileds:            */
/* uiMask,uiBytesNo,                                     */
/* It does not take into account Shift bits:            */
/* cShiftBits                                           */
/*------------------------------------------------------*/


int READ_AI_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short sRecordType,
    char        *chanName,
    unsigned int *iVal
   
    ){

    struct io_GPMM	  *descrp;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */

    volatile unsigned int   lVal = 0;
    volatile unsigned int   lValTemp = 0;
    
    PRINT_DEBUG("READ_AI_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);

    if(parm!=NULL)
      regNode=parm->pRegLocation;
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}

	if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	     status = regNode->fnct(card,signal,parm,sRecordType,REC_ENTER,chanName,(void *)&lVal);
	  }
	
	if ( (status & PROCESS_DATA) == RETURN_OK){
	  /* read data from memory location */
	/* this is the way how to access data in D32 mode i.e. intiger by integer */
	if(descrp->sDataMode == VMED32){
	  sDiv=signal>>2; /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned int);
	  PRINT_DEBUG("AI: VMED32 Read data  stared at %d offest %d\n",sDiv,sRem);
	  
	    lVal=((unsigned int *) regNode->pLocalAddr)[sDiv];
	    PRINT_DEBUG("VAL[%.2d]--------->\t  %.8X\n",sDiv,(unsigned int)lVal);

	    lVal=lVal>>(sRem<<3); /* sRem * 8 == sRem<<3 this is to speed up the execution*/
	    PRINT_DEBUG("VAL[%.2d] aligned->\t  %.8X\n",sDiv,(unsigned int)lVal);
	    
	    /* if number of bytes to read excced the scope 
               of the  regNode->pLocalAddr)[sDiv] then read regNode->pLocalAddr)[++sDiv] */

	    if( parm->uiBytesNo + sRem > sizeof(unsigned int)){

	      lValTemp=((unsigned int *) regNode->pLocalAddr)[++sDiv];
	      PRINT_DEBUG("VAL[%.2d]--------->%.8X\n",sDiv,(unsigned int)lValTemp); 
	      lValTemp=lValTemp<<((sizeof(unsigned int)-sRem)<<3);
	      PRINT_DEBUG("VAL[%.2d] aligned->%.8X\n",sDiv,(unsigned int)lValTemp);
	      lVal=lVal | lValTemp;
	      PRINT_DEBUG("VAL merged->%.8X\n",(unsigned int)lVal);
	    }
	
	} /* VMED32 */
	else if(descrp->sDataMode == VMED16){
	  sDiv=signal>>1; /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned short);
	  PRINT_DEBUG("AI: VMED16 Read data  stared at %d offest %d\n",sDiv,sRem);

	  lVal=0;
	  lVal=((unsigned short *) regNode->pLocalAddr)[sDiv];
	  if( parm->uiBytesNo + sRem > sizeof(unsigned short))
	  lVal |=  ((unsigned short *) regNode->pLocalAddr)[++sDiv]<<16;
	  

	  /*((volatile unsigned short *) & lVal)[1]=((unsigned short *) regNode->pLocalAddr)[sDiv];
	    if( parm->uiBytesNo + sRem > sizeof(unsigned short))
	    ((volatile unsigned short *) & lVal)[0]=((unsigned short *) regNode->pLocalAddr)[++sDiv];
	    else
	      ((volatile unsigned short *) & lVal)[0]= 0;
	  */
	    PRINT_DEBUG("VAL[%.2d]--------->\t  %.8X\n",sDiv,(unsigned int)lVal);
	 
	    lVal=lVal>>(sRem<<3); /* sRem * 8 == sRem<<3 this is to speed up the execution*/
	    PRINT_DEBUG("VAL[%.2d] aligned->\t  %.8X\n",sDiv,(unsigned int)lVal);
	    
	    /* if number of bytes to read excced the scope 
               of the  regNode->pLocalAddr)[sDiv] then read regNode->pLocalAddr)[++sDiv] */

	      if( parm->uiBytesNo + sRem > sizeof(unsigned int)){

		lValTemp=((unsigned short *) regNode->pLocalAddr)[++sDiv];
		PRINT_DEBUG("VAL[%.2d]--------->%.8X\n",sDiv,(unsigned int)lValTemp); 
		lValTemp=lValTemp<<((sizeof(unsigned int)-sRem)<<3);
		PRINT_DEBUG("VAL[%.2d] aligned->%.8X\n",sDiv,(unsigned int)lValTemp);
		lVal=lVal | lValTemp;
		PRINT_DEBUG("VAL merged->%.8X\n",(unsigned int)lVal);
	      }	  

	}/* VMED16 */
	else if(descrp->sDataMode == VMED8){
	  
	  PRINT_DEBUG("AI: VMED8 Read data  stared at %d\n",signal);
	  lVal=0;
	  lVal=((unsigned char *) regNode->pLocalAddr)[signal];
	  if( parm->uiBytesNo  > 1)
	    lVal |=  ((unsigned char *) regNode->pLocalAddr)[signal+1]<<8;
	  if( parm->uiBytesNo  > 2)
	    lVal |=  ((unsigned char *) regNode->pLocalAddr)[signal+2]<<16;
	  if( parm->uiBytesNo  > 3)
	    lVal |=  ((unsigned char *) regNode->pLocalAddr)[signal+3]<<24;

	  /*((volatile unsigned char*)&lVal)[3] = ((unsigned char *) regNode->pLocalAddr)[signal];
	  ((volatile unsigned char*)&lVal)[2] = ((unsigned char *) regNode->pLocalAddr)[signal+1];
	  ((volatile unsigned char*)&lVal)[1] = ((unsigned char *) regNode->pLocalAddr)[signal+2];
	  ((volatile unsigned char*)&lVal)[0] = ((unsigned char *) regNode->pLocalAddr)[signal+3];*/
	  PRINT_DEBUG("VAL raw    ->\t  %.8X\n",(unsigned int) lVal);	  

	}/* VMED8 */

	} /* (status & PROCESS_DATA) == RETURN_OK */
	
	if((status & ENTER_FUNC) == RETURN_OK){

	  if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	    status = regNode->fnct(card,signal,parm,sRecordType,REC_EXIT,chanName,(void *)&lVal);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */
	
	
	if((status & ERROR_OCCURED) == RETURN_OK){
	  PRINT_DEBUG("Before leaving READ_AI_GPMM  lVal %d\n",lVal);
	  *iVal=lVal;
	}
	else
	  {
	    printf("An error NR 0x%.4X  occured in AI:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }
	  
	
      } /* regNode != NULL */


    return(Ok);

}


/*------------------------------------------------------*/
/*               Analog Output write                    */
/* This function deals with the mask fileds:            */
/* uiMask,uiBytesNo                                      */
/* It does not take into account Shift bits:            */
/* cShiftBits                                           */
/*------------------------------------------------------*/



int WRITE_AO_GPMM(
    unsigned short  card,
    unsigned short  signal,
    struct recDesc *parm,
    unsigned short sRecordType,
    char           *chanName,
    unsigned int    iVal
    )
  {
    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
    volatile unsigned int  lVal = 0;
    unsigned int           lVal1= 0;
    volatile unsigned int  lValTemp = 0;
#ifdef DEBUG_SOFT
    unsigned int           lTmpTab[2]={0x12345678,0x9ABCDE3A}; /* TEST ONLY  */
#endif

    PRINT_DEBUG("WRITE_AO_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);
    PRINT_DEBUG("WRITE_AO_GPMM Value to be written 0x%.8X\n",iVal);
    if(parm!=NULL)
	  regNode=parm->pRegLocation;
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}

#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this lane should be deleted */
#endif
	 lVal=iVal;
	 if(regNode->fnct != NULL){
	     /* execut the user defined process function */
	     status = regNode->fnct(card,signal,parm,sRecordType,REC_ENTER,chanName,(void *)&lVal);
	   }

	 if ((status & PROCESS_DATA) == RETURN_OK){
	/* this is the way how to access the data in the A32 mode ie. integer by integer */
	   if(descrp->sDataMode == VMED32){

	     sDiv=signal>>2;        /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	     sRem=signal%sizeof(unsigned int);

	     PRINT_DEBUG("AO: VMED32 Read data started at %d offest %d\n",sDiv,sRem);

	     lValTemp=((unsigned int *) regNode->pLocalAddr)[sDiv];
	     PRINT_DEBUG("VAL[%.2d] read out>\t %.8X\n",sDiv,(unsigned int)lValTemp);

	     lValTemp=lValTemp & (~((parm->uiMask)<<(sRem<<3)));
	     PRINT_DEBUG("VAL[%.2d] masked-->\t %.8X\n",sDiv,(unsigned int)lValTemp);

	     lVal1=(lVal & parm->uiMask)<<(sRem<<3);
	     PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)lVal1);
	     lValTemp=lValTemp | lVal1;
	     PRINT_DEBUG("VAL[%.2d] merged-->\t %.8X\n",sDiv,(unsigned int)lValTemp);
	    
	     ((unsigned int *) regNode->pLocalAddr)[sDiv]=lValTemp;
	       
	     /* if number of bytes to read excced the scope of the  
		regNode->pLocalAddr)[sDiv] then read regNode->pLocalAddr)[++sDiv] */

	     if( parm->uiBytesNo + sRem > sizeof(unsigned int)){
	       lVal=(lVal & parm->uiMask)>>((sizeof(unsigned int)-sRem)<<3);              
	       PRINT_DEBUG("Value requested->%.8X\n",(unsigned int)lVal);
	       
	       lValTemp=((unsigned int *) regNode->pLocalAddr)[++sDiv]; /* from now on the sDiv = sDiv+1 */ 
	       PRINT_DEBUG("VAL[%.2d] read out>%.8X\n",sDiv,(unsigned int)lValTemp);

	       lValTemp=lValTemp & (~((parm->uiMask)>>((sizeof(unsigned int)-sRem)<<3)));
	       PRINT_DEBUG("VAL[%.2d] masked-->%.8X\n",sDiv,(unsigned int)lValTemp);

	       lValTemp=lValTemp | lVal;
	       PRINT_DEBUG("VAL[%.2d] merged-->%.8X\n",sDiv,(unsigned int)lValTemp);

	       ((unsigned int *) regNode->pLocalAddr)[sDiv]=lValTemp;
	      
	     }
	   }/* VMED32 mode */
	   else if(descrp->sDataMode == VMED16){
	     
             sDiv=signal>>1;        /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	     sRem=signal%sizeof(unsigned short);

	     PRINT_DEBUG("AO: VMED16 Read data started at %d offest %d\n",sDiv,sRem);

	     lValTemp  =0;
	     lValTemp  =  ((unsigned short *) regNode->pLocalAddr)[sDiv];
	     if( parm->uiBytesNo + sRem > sizeof(unsigned short))
	     lValTemp |=  ((unsigned short *) regNode->pLocalAddr)[sDiv+1]<<16;

             /* ((unsigned short *) lValTemp)[1] = ((unsigned short *) regNode->pLocalAddr)[sDiv];
		((unsigned short *) lValTemp)[0] = ((unsigned short *) regNode->pLocalAddr)[sDiv+1];*/

	     PRINT_DEBUG("VAL[%.2d] read out>\t %.8X\n",sDiv,(unsigned int)lValTemp);

	     lValTemp=lValTemp & (~((parm->uiMask)<<(sRem<<3)));
	     PRINT_DEBUG("VAL[%.2d] masked-->\t %.8X\n",sDiv,(unsigned int)lValTemp);

	     lVal1=(lVal & parm->uiMask)<<(sRem<<3);
	     PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)lVal1);
	     lValTemp=lValTemp | lVal1;
	     PRINT_DEBUG("VAL[%.2d] merged-->\t %.8X\n",sDiv,(unsigned int)lValTemp);
	    
	     
	     ((unsigned short *) regNode->pLocalAddr)[sDiv]  =(unsigned short )lValTemp;
	     if( parm->uiBytesNo + sRem > sizeof(unsigned short))
	     ((unsigned short *) regNode->pLocalAddr)[sDiv+1]=(unsigned short )lValTemp>>16;

	     /*((unsigned short *) regNode->pLocalAddr)[sDiv]=((unsigned short *) lValTemp)[1];
	       ((unsigned short *) regNode->pLocalAddr)[sDiv+1]=((unsigned short *) lValTemp)[0];*/

	     /* if number of bytes to read excced the scope of the  
		regNode->pLocalAddr)[sDiv] then read regNode->pLocalAddr)[++sDiv] */

	     if( parm->uiBytesNo + sRem > sizeof(unsigned int)){
	       lVal=(lVal & parm->uiMask)>>((sizeof(unsigned int)-sRem)<<3);              
	       PRINT_DEBUG("Value requested->%.8X\n",(unsigned int)lVal);
               sDiv = sDiv + 2;
	       lValTemp=((unsigned short *) regNode->pLocalAddr)[sDiv]; /* from now on the sDiv = sDiv+2 */ 
	       PRINT_DEBUG("VAL[%.2d] read out>%.8X\n",sDiv,(unsigned int)lValTemp);

	       lValTemp=lValTemp & (~((parm->uiMask)>>((sizeof(unsigned int)-sRem)<<3)));
	       PRINT_DEBUG("VAL[%.2d] masked-->%.8X\n",sDiv,(unsigned int)lValTemp);

	       lValTemp=lValTemp | lVal;
	       PRINT_DEBUG("VAL[%.2d] merged-->%.8X\n",sDiv,(unsigned int)lValTemp);

	       ((unsigned int *) regNode->pLocalAddr)[sDiv]=lValTemp;
	      
	     }
	     
	   }/* VMED16 mode */

	   else if(descrp->sDataMode == VMED8){
	     PRINT_DEBUG("AO: VMED8: Read data  started at %d \n",signal);
	     lValTemp=0;
	     lValTemp=((unsigned char *) regNode->pLocalAddr)[signal];
	     if( parm->uiBytesNo  > 1)
	       lValTemp |=  ((unsigned char *) regNode->pLocalAddr)[signal+1]<<8;
	     if( parm->uiBytesNo  > 2)
	       lValTemp |=  ((unsigned char *) regNode->pLocalAddr)[signal+2]<<16;
	     if( parm->uiBytesNo  > 3)
	       lValTemp |=  ((unsigned char *) regNode->pLocalAddr)[signal+3]<<24;

	     

	     PRINT_DEBUG("VAL--------->\t %.8X\n",(unsigned int) lValTemp);
	     lValTemp=lValTemp & (~(parm->uiMask));
	     PRINT_DEBUG("VAL masked-->\t %.8X\n",(unsigned int)lValTemp);
	     lVal1=(lVal & parm->uiMask);
	     PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)lVal1);
	     lVal=lValTemp | lVal1;
	     PRINT_DEBUG("VAL merged-->\t %.8X\n",(unsigned int)lVal);
	     

	     ((unsigned char *) regNode->pLocalAddr)[signal] =      lVal;
	     if( parm->uiBytesNo  > 1)
	       ((unsigned char *) regNode->pLocalAddr)[signal+1] =  lVal >> 8;
	     if( parm->uiBytesNo  > 2)
	       ((unsigned char *) regNode->pLocalAddr)[signal+1] =  lVal >> 16;
	     if( parm->uiBytesNo  > 3)
	       ((unsigned char *) regNode->pLocalAddr)[signal+1] =  lVal >> 24;
    
	   }/* VMED8 mode */
	 }
	 
	 if((status & ENTER_FUNC) == RETURN_OK){

	  if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	    status = regNode->fnct(card,signal,parm,sRecordType,REC_EXIT,chanName,(void *)&lVal);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */
	 
	 if((status & ERROR_OCCURED) != RETURN_OK){
	    printf("An error NR 0x%.4X  occured in AO:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }


      }/* regNode != NULL */
     

    return(Ok);
}

/*------------------------------------------------------*/
/*               Binary  Input raed                     */
/* This function deals with the:                        */
/* Shift bits:                                          */
/* cShiftBits                                           */
/*------------------------------------------------------*/


int READ_BI_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    char  *chanName,
    unsigned short *val
    ){

    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
    volatile unsigned int  iVal = 0;
    
    unsigned short sShiftBits=0;

#ifdef DEBUG_SOFT
    unsigned int           lTmpTab[2]={0x12345678,0x9ABCDE3A}; /* TEST ONLY this lane should be deleted */
#endif

    PRINT_DEBUG("READ_BI_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);

    if(parm!=NULL){
      regNode=parm->pRegLocation;
      sShiftBits=parm->cShiftBits;
    }
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this lane should be deleted */
#endif

	if(regNode->fnct != NULL){
	  /* execut the user defined process function */	     
	  status = regNode->fnct(card,signal,parm,REC_BI,REC_ENTER,chanName,(void *)&iVal);
	}

	if ( (status & PROCESS_DATA) == RETURN_OK){
	/* this is the way how to access data in D32 mode i.e. intiger by integer */
	if(descrp->sDataMode == VMED32){
	  sDiv=signal>>2; /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned int);
          PRINT_DEBUG("BI: VMED32: Read data  started at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);

	  iVal=((unsigned int *) regNode->pLocalAddr)[sDiv];
	  PRINT_DEBUG("VAL[%.2d]--------->\t  %.8X\n",sDiv,(unsigned int) iVal);

	  iVal=iVal>>((sRem<<3)+sShiftBits); /* sRem * 8 bits == sRem<<3 this is to speed up the execution*/
	  PRINT_DEBUG("VAL[%.2d] aligned->\t  %.8X\n",sDiv,(unsigned int) iVal);	  

	  iVal=iVal & 0x1;
	  PRINT_DEBUG("VAL[%.2d] masked ->\t  %.8X\n",sDiv,(unsigned int) iVal);

	  
	}/* VMED32 */ 
	else if(descrp->sDataMode == VMED16){
	  iVal=0;
	  sDiv=signal>>1; /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned short);

          PRINT_DEBUG("BI: VMED16: Read data  started at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);

	  iVal  = 0;
	  iVal  =  ((unsigned short *) regNode->pLocalAddr)[sDiv];
	  if(((sRem<<3)+sShiftBits) > 16 )
	    iVal |=  ((unsigned short *) regNode->pLocalAddr)[sDiv+1]<<16;

	  /* ((volatile unsigned short*)&iVal)[1] = ((unsigned short *) regNode->pLocalAddr)[sDiv];

	  if(((sRem<<3)+sShiftBits) > 16 ) 
	    ((volatile unsigned short*)&iVal)[0] = ((unsigned short *) regNode->pLocalAddr)[sDiv+1];
	  else
	  ((volatile unsigned short*)&iVal)[0] = 0; */
	  
	 
	  PRINT_DEBUG("VAL raw    ->\t  %.8X\n",(unsigned int) iVal);
	  iVal=iVal>>sShiftBits ; 
	  PRINT_DEBUG("VAL aligned->\t  %.8X\n",(unsigned int) iVal);	  
	  iVal=iVal & 0x1;
	  PRINT_DEBUG("VAL masked ->\t  %.8X\n",(unsigned int) iVal);

	}/* VMED16 */
	else if(descrp->sDataMode == VMED8){
	  
	  PRINT_DEBUG("BI: VMED8 Read data  started at %d  shft bits %d\n",signal,sShiftBits);
	  iVal=0;

	  iVal  =  ((unsigned char *) regNode->pLocalAddr)[signal];
	  iVal |=  ((unsigned char *) regNode->pLocalAddr)[signal+1]<<8;

	  /* ((volatile unsigned char*)&iVal)[3] = ((unsigned char *) regNode->pLocalAddr)[signal];
	  ((volatile unsigned char*)&iVal)[2] = ((unsigned char *) regNode->pLocalAddr)[signal+1];
	  ((volatile unsigned char*)&iVal)[1] = 0;
	  ((volatile unsigned char*)&iVal)[0] = 0; */
	  PRINT_DEBUG("VAL raw    ->\t  %.8X\n",(unsigned int) iVal);
	  iVal=iVal>>sShiftBits ; 
	  PRINT_DEBUG("VAL aligned->\t  %.8X\n",(unsigned int) iVal);	  
	  iVal=iVal & 0x1;
	  PRINT_DEBUG("VAL masked ->\t  %.8X\n",(unsigned int) iVal);

	}/* VMED8 */
	} /* (status & PROCESS_DATA) == RETURN_OK */

	if((status & ENTER_FUNC) == RETURN_OK){
	  if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    status = regNode->fnct(card,signal,parm,REC_BI,REC_EXIT,chanName,(void *)&iVal);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */
	
	if((status & ERROR_OCCURED) == RETURN_OK){
	  PRINT_DEBUG("Before leaving  READ_BI GPMM iVal %d\n",iVal);
	  *val=(unsigned short) iVal;
	}
	else
	  {
	    printf("An error NR 0x%.4X  occured in BI:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }

      }/* regNode != NULL */

    return(Ok);

}

/*------------------------------------------------------*/
/*               Binary OUTput write                    */
/* This function deals with:                            */
/* Shift bits:                                          */
/* cShiftBits                                           */
/*------------------------------------------------------*/

int WRITE_BO_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    char  *chanName,
    unsigned short rVal
    )
  {
    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
   
    volatile unsigned int iValTemp=0;
    unsigned int          iVal1=0;
    unsigned short sShiftBits=0;

#ifdef DEBUG_SOFT
    unsigned int           lTmpTab[2]={0x12345678,0x9ABCDE3A}; /* TEST ONLY this lane should be deleted */
#endif

    PRINT_DEBUG("WRITE_BO_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);
    PRINT_DEBUG("WRITE_BO_GPMM Value to be written 0x%.4X\n",rVal);

    if(parm!=NULL){
	  regNode=parm->pRegLocation;
	  sShiftBits=parm->cShiftBits;
    }
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this lane should be deleted */
#endif
    /* if the desired record was found then set the value according to the signal number "iRefNr" */

	 if(regNode->fnct != NULL){
	     /* execut the user defined process function */
	     status = regNode->fnct(card,signal,parm,REC_BO,REC_ENTER,chanName,(void *)&rVal);
	   }
	 if ((status & PROCESS_DATA) == RETURN_OK){
	/* this is the way how to access the data in the A32 mode ie. integer by integer */
	 if(descrp->sDataMode == VMED32){

	   sDiv=signal>>2;        /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	   sRem=signal%sizeof(unsigned int);

	   PRINT_DEBUG("BO: VMED32 Read data  stared at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);
	   iValTemp=((unsigned int *) regNode->pLocalAddr)[sDiv];

	   PRINT_DEBUG("VAL[%.2d]--------->\t %.8X\n",sDiv,(unsigned int) iValTemp);

	   iValTemp=iValTemp & (~((0x1)<<((sRem<<3)+sShiftBits)));
	   PRINT_DEBUG("VAL[%.2d] masked-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   iVal1=rVal;
	   iVal1=(iVal1 & 0x1)<<((sRem<<3)+sShiftBits);
	   PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)iVal1);
	   iValTemp=iValTemp | iVal1;
	   PRINT_DEBUG("VAL[%.2d] merged-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   ((unsigned int *) regNode->pLocalAddr)[sDiv]=iValTemp;

	   

	 } /* VMED32 mode */
	 else if(descrp->sDataMode == VMED16){

	   sDiv=signal>>1;        /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	   sRem=signal%sizeof(unsigned short);

	   PRINT_DEBUG("BO: VMED16 Read data  stared at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);

	   iValTemp  = 0;
	   iValTemp  = ((unsigned short *) regNode->pLocalAddr)[sDiv];
	   if(((sRem<<3)+sShiftBits) >= 16 )
	     iValTemp |=  ((unsigned short *) regNode->pLocalAddr)[sDiv+1]<<16;

	   /*  ((volatile unsigned short*)&iValTemp)[1] = ((unsigned short *) regNode->pLocalAddr)[sDiv];
	  if(((sRem<<3)+sShiftBits) > 16 ) 
	    ((volatile unsigned short*)&iValTemp)[0] = ((unsigned short *) regNode->pLocalAddr)[sDiv+1];
	  else
	  ((volatile unsigned short*)&iValTemp)[0] = 0;*/

	   PRINT_DEBUG("VAL[%.2d]--------->\t %.8X\n",sDiv,(unsigned int) iValTemp);

	   iValTemp=iValTemp & (~((0x1)<<((sRem<<3)+sShiftBits)));
	   PRINT_DEBUG("VAL[%.2d] masked-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   iVal1=rVal;
	   iVal1=(iVal1 & 0x1)<<((sRem<<3)+sShiftBits);
	   PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)iVal1);
	   iValTemp=iValTemp | iVal1;
	   PRINT_DEBUG("VAL[%.2d] merged-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   ((unsigned short *) regNode->pLocalAddr)[sDiv]     =iValTemp;
	    if(((sRem<<3)+sShiftBits) >= 16 )
	      ((unsigned short *) regNode->pLocalAddr)[sDiv+1]=iValTemp>>16;

	   /* ((unsigned short *) regNode->pLocalAddr)[sDiv]=((volatile unsigned short*)&iValTemp)[1];
	   if(((sRem<<3)+sShiftBits) > 16 ) 
	     ((unsigned short *) regNode->pLocalAddr)[sDiv+1]= ((volatile unsigned short*)&iValTemp)[0]; */

	 } /* VMED16 mode */

	 else if(descrp->sDataMode == VMED8){

	   PRINT_DEBUG("VMED8: Read data at the location stared at %d  shift bits %d\n",signal,sShiftBits);
	   iValTemp  = 0;
	   iValTemp  = ((unsigned char *) regNode->pLocalAddr)[sDiv];
	   if(sShiftBits > 0x7 )
	     iValTemp |=  ((unsigned char *) regNode->pLocalAddr)[sDiv+1]<<8;

	   /* ((volatile unsigned char *)&iValTemp)[3]=((unsigned char *) regNode->pLocalAddr)[signal];
	   if(sShiftBits > 0x7)
	     ((volatile unsigned char *)&iValTemp)[2]=((unsigned char *) regNode->pLocalAddr)[signal+1];
	   else
	     ((volatile unsigned char *)&iValTemp)[2]= 0;
	   ((volatile unsigned char *)&iValTemp)[1]=0;
	   ((volatile unsigned char *)&iValTemp)[0]=0;*/


	   PRINT_DEBUG("VAL --------->\t %.8X\n",(unsigned int) iValTemp);
	   iValTemp=iValTemp & (~((0x1)<<sShiftBits));
	   PRINT_DEBUG("VAL  masked-->\t %.8X\n",(unsigned int)iValTemp);
	   iVal1=rVal;
	   iVal1=(iVal1 & 0x1)<<(sShiftBits);
	   PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)iVal1);
	   iValTemp=iValTemp | iVal1;
	   PRINT_DEBUG("VAL  merged-->\t %.8X\n",(unsigned int)iValTemp);


	   ((unsigned char *) regNode->pLocalAddr)[signal]     =iValTemp;
	    if(sShiftBits>0x7)
	      ((unsigned char *) regNode->pLocalAddr)[signal+1]=iValTemp>>8;

	    /* ((unsigned char *) regNode->pLocalAddr)[signal]    =((volatile unsigned char *)&iValTemp)[3];
	   if(sShiftBits>0x7)
	   ((unsigned char *) regNode->pLocalAddr)[signal+1]=((volatile unsigned char *)&iValTemp)[2]; */


	 } /* VMED8 mode */
	 }/* (status & PROCESS_DATA) == RETURN_OK */

	 if((status & ENTER_FUNC) == RETURN_OK){

	  if(regNode->fnct != NULL){
	    status = regNode->fnct(card,signal,parm,REC_BO,REC_EXIT,chanName,(void *)&rVal);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */
	 
	 if((status & ERROR_OCCURED) != RETURN_OK){
	    printf("An error NR 0x%.4X  occured in BO:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }

      }/* regNode != NULL */
    return(Ok);
}

/*------------------------------------------------------*/
/*               MBBI Direct read                       */
/* This function deals with the mask fileds:            */
/* uiMask,cMaskBitNr                                    */
/* and Shift bits:                                      */
/* cShiftBits                                           */
/*------------------------------------------------------*/

int READ_MBBI_DIRRECT_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short sRecordType,
    char  *chanName,
    unsigned short *val
    ){

    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
    volatile unsigned int   iVal = 0, iValTmp=0;
    
    unsigned short sShiftBits=0;

#ifdef DEBUG_SOFT
    unsigned int           lTmpTab[2]={0x12345678,0x9ABCDE3A}; /* TEST ONLY this lane should be deleted */
#endif

    PRINT_DEBUG("READ_MBBI_D_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);

    if(parm!=NULL){
      regNode=parm->pRegLocation;
      sShiftBits=parm->cShiftBits;
    }
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this lane should be deleted */
#endif

	
	if(regNode->fnct != NULL){
	  /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	  status = regNode->fnct(card,signal,parm,sRecordType,REC_ENTER,chanName,(void *)&iVal); 
	}
	
	if ( (status & PROCESS_DATA) == RETURN_OK){

	/* this is the way how to access data in D32 mode i.e. intiger by integer */
	if(descrp->sDataMode == VMED32){
	  sDiv=signal>>2; /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned int);

	  PRINT_DEBUG("MBBI D: VMED32 Read data  stared at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);

	  iVal=((unsigned int *) regNode->pLocalAddr)[sDiv];
	  PRINT_DEBUG("VAL[%.2d]--------->\t  %.8X\n",sDiv,(unsigned int) iVal);

	  iVal=iVal>>((sRem<<3)+sShiftBits); /* sRem * 8 bits == sRem<<3 this is to speed up the execution*/
	  PRINT_DEBUG("VAL[%.2d] aligned->\t  %.8X\n",sDiv,(unsigned int) iVal);
	  
	  /* read the second unsigned int ? */
	  if( ((parm->cMaskBitNr + sShiftBits) + (sRem<<3)) > 32)   /* 32  = sizeof(unsigned int)*8 bits */
	    {
	      iValTmp=((unsigned int *) regNode->pLocalAddr)[++sDiv];
	      PRINT_DEBUG("VAL[%.2d]--------->%.8X\n",sDiv,(unsigned int) iValTmp);

	      iValTmp=iValTmp<<(((sizeof(unsigned int)-sRem)<<3)-sShiftBits);
	      PRINT_DEBUG("VAL[%.2d] aligned->%.8X\n",sDiv,(unsigned int)iValTmp);
	      
	      iVal= iVal | iValTmp;
	      PRINT_DEBUG("VAL merged->%.4X\n",(unsigned short) iVal);
	    }
	  
	}/* VMED32 */ 
	else if(descrp->sDataMode == VMED16){
	  sDiv=signal>>1; /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned short);

	  PRINT_DEBUG("MBBI D: VMED16 Read data  stared at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);
	  
	  iVal=0;
	  iVal=((unsigned short *) regNode->pLocalAddr)[sDiv];
	  if(((sRem<<3)+sShiftBits + parm->cMaskBitNr) > 16 ) /* 16 means sizeof(short) */
	  iVal |=  ((unsigned short *) regNode->pLocalAddr)[sDiv+1]<<16;


	  /*((volatile unsigned short*)&iVal)[1] = ((unsigned short *) regNode->pLocalAddr)[sDiv];

	  if(((sRem<<3)+sShiftBits + parm->cMaskBitNr) > 16 ) 
	    ((volatile unsigned short*)&iVal)[0] = ((unsigned short *) regNode->pLocalAddr)[sDiv+1];
	  else
	  ((volatile unsigned short*)&iVal)[0] = 0;*/


	  PRINT_DEBUG("VAL[%.2d]--------->\t  %.8X\n",sDiv,(unsigned int) iVal);

	  iVal=iVal>>((sRem<<3)+sShiftBits); /* sRem * 8 bits == sRem<<3 this is to speed up the execution*/
	  PRINT_DEBUG("VAL[%.2d] aligned->\t  %.8X\n",sDiv,(unsigned int) iVal);
	  
	  /* read the second unsigned int ? */
	  if( ((parm->cMaskBitNr + sShiftBits) + (sRem<<3)) > 32)   /* 32  = sizeof(unsigned int)*8 bits */
	    {
	      sDiv = sDiv + 2;
	      iValTmp=((unsigned short *) regNode->pLocalAddr)[sDiv];
	      PRINT_DEBUG("VAL[%.2d]--------->%.8X\n",sDiv,(unsigned int) iValTmp);

	      iValTmp=iValTmp<<(((sizeof(unsigned int)-sRem)<<3)-sShiftBits);
	      PRINT_DEBUG("VAL[%.2d] aligned->%.8X\n",sDiv,(unsigned int)iValTmp);
	      
	      iVal= iVal | iValTmp;
	      PRINT_DEBUG("VAL merged->%.4X\n",(unsigned short) iVal);
	    }
	  
	}/* VMED16 */ 

	else if(descrp->sDataMode == VMED8){
	 
	  PRINT_DEBUG("MBBI D: VMED8 Read data  stared at %d  shft bits %d\n",signal,sShiftBits);

	  iVal  = 0;
	  iVal  =  ((unsigned char *) regNode->pLocalAddr)[signal];
	  iVal |=  ((unsigned char *) regNode->pLocalAddr)[signal+1]<<8;
	  if((sShiftBits + parm->cMaskBitNr) > 16 ) /* 16 means sizeof(short) */
	  iVal |=  ((unsigned char *) regNode->pLocalAddr)[signal+2]<<16;

	  /* ((volatile unsigned char*)&iVal)[3] = ((unsigned char *) regNode->pLocalAddr)[signal];
	  ((volatile unsigned char*)&iVal)[2] = ((unsigned char *) regNode->pLocalAddr)[signal+1];
	  ((volatile unsigned char*)&iVal)[1] = 0;
	  ((volatile unsigned char*)&iVal)[0] = 0; */

	  PRINT_DEBUG("VAL raw    ->\t  %.8X\n",(unsigned int) iVal);
	  iVal=iVal>>sShiftBits ; 
	  PRINT_DEBUG("VAL aligned->\t  %.8X\n",(unsigned int) iVal);	  

	}/* VMED8 */
	} /* (status & PROCESS_DATA) == RETURN_OK */

	if((status & ENTER_FUNC) == RETURN_OK){

	  if(regNode->fnct != NULL){	    
	    status = regNode->fnct(card,signal,parm,sRecordType,REC_EXIT,chanName,(void *)&iVal);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */

	if((status & ERROR_OCCURED) == RETURN_OK){
	  PRINT_DEBUG("Before leaving READ_MBBI_D_GPMM  iVal %d\n",iVal);
	  *val=(unsigned short) iVal;
	}
	else
	  {
	    printf("An error NR 0x%.4X  occured in MBBI_D:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }

      }/* regNode != NULL */

    return(Ok);

}

/*------------------------------------------------------*/
/*               MBBO Direct   write                    */
/* This function deals with the mask fileds:            */
/* uiMask,cMaskBitNr                                    */
/* and Shift bits:                                      */
/* cShiftBits                                           */
/*------------------------------------------------------*/

int WRITE_MBBO_DIRRECT_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short sRecordType,
    char  *chanName,
    unsigned short rVal
    )
  {
    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
   
    volatile unsigned int iValTemp=0;
    unsigned int          iVal1=0;
    unsigned short sShiftBits=0;

#ifdef DEBUG_SOFT
    unsigned int           lTmpTab[2]={0x12345678,0x9ABCDE3A}; /* TEST ONLY this lane should be deleted */
#endif

    PRINT_DEBUG("WRITE_MBBO_D_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);
    PRINT_DEBUG("WRITE_MBBO_D_GPMM Value to be written 0x%.4X\n",rVal);

    if(parm!=NULL){
	  regNode=parm->pRegLocation;
	  sShiftBits=parm->cShiftBits;
    }
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this lane should be deleted */
#endif
    /* if the desired record was found then set the value according to the signal number "iRefNr" */

	 if(regNode->fnct != NULL){
	     /* execut the user defined process function */
	     status = regNode->fnct(card,signal,parm,sRecordType,REC_ENTER,chanName,(void *)&rVal);
	   }
	 if ((status & PROCESS_DATA) == RETURN_OK){
	/* this is the way how to access the data in the A32 mode ie. integer by integer */
	 if(descrp->sDataMode == VMED32){

	   sDiv=signal>>2;        /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	   sRem=signal%sizeof(unsigned int);

	   PRINT_DEBUG("MBBO D: VMED32 Read data  stared at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);
	   
	   iValTemp=((unsigned int *) regNode->pLocalAddr)[sDiv];
	   PRINT_DEBUG("VAL[%.2d]--------->\t %.8X\n",sDiv,(unsigned int) iValTemp);

	   iValTemp=iValTemp & (~((parm->uiMask)<<((sRem<<3)+sShiftBits)));
	   PRINT_DEBUG("VAL[%.2d] masked-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   iVal1=rVal;
	   iVal1=(iVal1 & parm->uiMask)<<((sRem<<3)+sShiftBits);
	   PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)iVal1);
	   iValTemp=iValTemp | iVal1;
	   PRINT_DEBUG("VAL[%.2d] merged-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   ((unsigned int *) regNode->pLocalAddr)[sDiv]=iValTemp;


	   if( parm->cMaskBitNr + sShiftBits + (sRem<<3) > 32 )/* 32  = sizeof(unsigned int)*8 bits */
	     {
	       iVal1=(unsigned int)(rVal & parm->uiMask)>>(((sizeof(unsigned int)-sRem)<<3)-sShiftBits);      
	       PRINT_DEBUG("Value requested->%.8X\n",(unsigned int)iVal1);

	       iValTemp=((unsigned int *) regNode->pLocalAddr)[++sDiv]; /* from now on the sDiv = sDiv+1 */ 
	       PRINT_DEBUG("VAL[%.2d] read out>%.8X\n",sDiv,(unsigned int)iValTemp);
	       
	       iValTemp=iValTemp & (~((parm->uiMask)>>(((sizeof(unsigned int)-sRem)<<3)-sShiftBits)));
	       PRINT_DEBUG("VAL[%.2d] masked-->%.8X\n",sDiv,(unsigned int)iValTemp);

	       iValTemp=iValTemp | iVal1;
	       PRINT_DEBUG("VAL[%.2d] merged-->%.8X\n",sDiv,(unsigned int)iValTemp);

	       ((unsigned int *) regNode->pLocalAddr)[sDiv]=iValTemp;
	      
	     }

	 } /* VMED32 mode */

	 else if(descrp->sDataMode == VMED16){

	   sDiv=signal>>1;        /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	   sRem=signal%sizeof(unsigned short);

	   
	   PRINT_DEBUG("MBBO D: VMED16 Read data  stared at %d offest %d shft bits %d\n",sDiv,sRem,sShiftBits);
	   iValTemp=0;
	   iValTemp=((unsigned short *) regNode->pLocalAddr)[sDiv];
	   if(((sRem<<3)+sShiftBits + parm->cMaskBitNr) > 16 ) /* 16 means sizeof(short) */
	     iValTemp |=  ((unsigned short *) regNode->pLocalAddr)[sDiv+1]<<16;
	   
	   /* ((volatile unsigned short*)&iValTemp)[1] = ((unsigned short *) regNode->pLocalAddr)[sDiv];

	  if(((sRem<<3)+sShiftBits + parm->cMaskBitNr) > 16 ) 
	    ((volatile unsigned short*)&iValTemp)[0] = ((unsigned short *) regNode->pLocalAddr)[sDiv+1];
	  else
	    ((volatile unsigned short*)&iValTemp)[0] = 0; */

	   PRINT_DEBUG("VAL[%.2d]--------->\t %.8X\n",sDiv,(unsigned int) iValTemp);

	   iValTemp=iValTemp & (~((parm->uiMask)<<((sRem<<3)+sShiftBits)));
	   PRINT_DEBUG("VAL[%.2d] masked-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   iVal1=rVal;
	   iVal1=(iVal1 & parm->uiMask)<<((sRem<<3)+sShiftBits);
	   PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)iVal1);
	   iValTemp=iValTemp | iVal1;
	   PRINT_DEBUG("VAL[%.2d] merged-->\t %.8X\n",sDiv,(unsigned int)iValTemp);

	   ((unsigned short *) regNode->pLocalAddr)[sDiv]    =  iValTemp;
	   if(((sRem<<3)+sShiftBits + parm->cMaskBitNr) > 16 ) /* 16 means sizeof(short) */
	    ((unsigned short *) regNode->pLocalAddr)[sDiv+1] =  iValTemp >> 16;

	   /* ((unsigned short *) regNode->pLocalAddr)[sDiv] = ((volatile unsigned short*)&iValTemp)[1];

	   if(((sRem<<3)+sShiftBits + parm->cMaskBitNr) > 16 ) 
	   ((unsigned short *) regNode->pLocalAddr)[sDiv+1] = ((volatile unsigned short*)&iValTemp)[0];*/
	  

	   if( parm->cMaskBitNr + sShiftBits + (sRem<<3) > 32 )/* 32  = sizeof(unsigned int)*8 bits */
	     {
	       iVal1=(unsigned int)(rVal & parm->uiMask)>>(((sizeof(unsigned int)-sRem)<<3)-sShiftBits);      
	       PRINT_DEBUG("Value requested->%.8X\n",(unsigned int)iVal1);
	       sDiv = sDiv + 2;
	       iValTemp=((unsigned short *) regNode->pLocalAddr)[sDiv]; /* from now on the sDiv = sDiv+2 */ 
	       PRINT_DEBUG("VAL[%.2d] read out>%.8X\n",sDiv,(unsigned int)iValTemp);
	       
	       iValTemp=iValTemp & (~((parm->uiMask)>>(((sizeof(unsigned int)-sRem)<<3)-sShiftBits)));
	       PRINT_DEBUG("VAL[%.2d] masked-->%.8X\n",sDiv,(unsigned int)iValTemp);

	       iValTemp=iValTemp | iVal1;
	       PRINT_DEBUG("VAL[%.2d] merged-->%.8X\n",sDiv,(unsigned int)iValTemp);

	       ((unsigned short *) regNode->pLocalAddr)[sDiv]=iValTemp;
	      
	     }

	 } /* VMED16 mode */

	 else if(descrp->sDataMode == VMED8){
	   PRINT_DEBUG("VMED8: Read data at the location stared at %d shift bits %d\n",signal,sShiftBits);

	  iValTemp  = 0;
	  iValTemp  =  ((unsigned char *) regNode->pLocalAddr)[signal];
	  if((sShiftBits + parm->cMaskBitNr) > 0x8 ) 
	  iValTemp |=  ((unsigned char *) regNode->pLocalAddr)[signal+1]<<8;
 	  if((sShiftBits + parm->cMaskBitNr) > 16 ) 
	  iValTemp |=  ((unsigned char *) regNode->pLocalAddr)[signal+2]<<16;

	   PRINT_DEBUG("VAL--------->\t %.8X\n",(unsigned int) iValTemp);
	   
	   iValTemp=iValTemp & (~((parm->uiMask)<<(sShiftBits)));
	   PRINT_DEBUG("VAL masked-->\t %.8X\n",(unsigned int)iValTemp);
	   
	   iVal1=rVal;
	   iVal1=(iVal1 & parm->uiMask)<<(sShiftBits);
	   PRINT_DEBUG("Value requested->\t %.8X\n",(unsigned int)iVal1);
	   iValTemp=iValTemp | iVal1;
	   PRINT_DEBUG("VAL merged-->\t %.8X\n",(unsigned int)iValTemp);


	  ((unsigned char *) regNode->pLocalAddr)[signal] = iValTemp   ;
	  if((sShiftBits + parm->cMaskBitNr) > 0x8 ) 
	  ((unsigned char *) regNode->pLocalAddr)[signal+1] = iValTemp >>8;
	  if((sShiftBits + parm->cMaskBitNr) > 16 ) 
	  ((unsigned char *) regNode->pLocalAddr)[signal+2] = iValTemp >>16;

	     
	 }/* VMED8 */
	 }/* (status & PROCESS_DATA) == RETURN_OK */

	 
	 if((status & ENTER_FUNC) == RETURN_OK){

	  if(regNode->fnct != NULL){
	    status = regNode->fnct(card,signal,parm,sRecordType,REC_EXIT,chanName,(void *)&rVal);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */
	 
	 if((status & ERROR_OCCURED) != RETURN_OK){
	    printf("An error NR 0x%.4X  occured in MBBO_D:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }

      }/* regNode != NULL */
    return(Ok);
}

/*------------------------------------------------------*/
/*               STR IN     read                        */
/*------------------------------------------------------*/

int READ_STR_IN_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    char  *chanName,
    char *val
    ){

    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
    volatile unsigned int   iVal = 0;
    unsigned short         sX=0;

#ifdef DEBUG_SOFT    
    char           lTmpTab[40]="ALA ma Asa a Ola ma kota."; /* TEST ONLY this line should be deleted */
#endif

    PRINT_DEBUG("READ_STR_IN_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);

    if(parm!=NULL)
      regNode=parm->pRegLocation;
 
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this line should be deleted */
#endif
	if(regNode->fnct != NULL){
	  /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	  status = regNode->fnct(card,signal,parm,REC_STR_IN,REC_ENTER,chanName,(void *)val);
	}
	
	if ( (status & PROCESS_DATA) == RETURN_OK){
	

	/* this is the way how to access data in D32 mode i.e. intiger by integer */
	if(descrp->sDataMode == VMED32){
	  sDiv=signal>>2; /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned int);

	  PRINT_DEBUG("Read data at the location stared at %d offest %d \n",sDiv,sRem);

	  if(sRem)
	    iVal=((unsigned int *) regNode->pLocalAddr)[sDiv++];

	  for(sX=0;sX<parm->uiBytesNo;sX++){
	    if(!sRem)
	      iVal=((unsigned int *) regNode->pLocalAddr)[sDiv++];
	    val[sX]=((char *)(&iVal))[sRem++];
	    sRem=sRem%sizeof(unsigned int);
	  }

	  
	  PRINT_DEBUG("STR=>  %s\n",val);

	  
	}/* VMED32 */ 
	else if(descrp->sDataMode == VMED16){
	  sDiv=signal>>1; /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned short);

	  PRINT_DEBUG("Read data at the location stared at %d offest %d \n",sDiv,sRem);

	  if(sRem)
	    iVal=((unsigned short *) regNode->pLocalAddr)[sDiv++];

	  for(sX=0;sX<parm->uiBytesNo;sX++){
	    if(!sRem)
	      iVal=((unsigned short *) regNode->pLocalAddr)[sDiv++];
	    val[sX]=((char *)(&iVal))[sRem++];
	    sRem=sRem%sizeof(unsigned short);
	  }

	  PRINT_DEBUG("STR=>  %s\n",val);
	  
	}/* VMED16 */ 


	else if(descrp->sDataMode == VMED8){
	  sDiv=signal;
	  for(sX=0;sX<parm->uiBytesNo;sX++){
	  ((volatile char *)val)[sX] = ((unsigned char *) regNode->pLocalAddr)[sDiv++];
	  }
	  PRINT_DEBUG("STR=>  %s\n",val);


	}/* VMED8 */
	}/* (status & PROCESS_DATA) == RETURN_OK) */


	if((status & ERROR_OCCURED) == RETURN_OK){
	  if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    status = regNode->fnct(card,signal,parm,REC_STR_IN,REC_EXIT,chanName,(void *)val);
	  }
	}/* (status & ENTER_FUNC) == RETURN_OK */

	if((status & ERROR_OCCURED) == RETURN_OK){
	  PRINT_DEBUG("Before leaving STR_IN_GPMM\n");
	}
	else
	  {
	    printf("An error NR 0x%.4X  occured in STR_IN:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }

      }/* regNode != NULL */

    return(Ok);

}



/*------------------------------------------------------*/
/*               WF  WRITE                              */
/* cShiftBits refers to the number of bytes per element */
/* ex: short -> 2 bytes                                 */
/*------------------------------------------------------*/

int WRITE_WF_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    char  *chanName,
    void  *pWf
    ){

    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
    /* volatile unsigned int   iVal = 0; */
    unsigned int           uiX=0;
    unsigned int           uiLimit=0;
    volatile unsigned int  iValTemp=0;
    void                   *pWfTemp=NULL;
    void                   *pWfDest=NULL;

#ifdef DEBUG_SOFT    
    char           lTmpTab[40]="ALA ma Asa a Ola ma kota."; /* TEST ONLY this line should be deleted */
#endif

    PRINT_DEBUG("WRITE_WF_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);
    if(pWf == NULL){
      printf("WF: wf points to NULL\n");
      return (-1);
    }

    if(parm!=NULL)
      regNode=parm->pRegLocation;
 
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this line should be deleted */
#endif
	

	if(regNode->fnct != NULL){
	  /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	   status = regNode->fnct(card,signal,parm,REC_WF,REC_ENTER,chanName,(void *)pWf);
	}

	/* this is the way how to access data in D32 mode i.e. intiger by integer */
	if ( (status & PROCESS_DATA) == RETURN_OK){
	if(descrp->sDataMode == VMED32){
	  sDiv=signal>>2; /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned int);

	  PRINT_DEBUG("WF: VMED32 Write data to %d offest %d \n",sDiv,sRem);

	  uiLimit=(parm->uiBytesNo)/sizeof(int);
	  PRINT_DEBUG("D32 uiLimit %d\tbytes %d\n",uiLimit,parm->uiBytesNo);
	  pWfTemp=pWf;
          /* -------------- Reminder is different then 0 part 1-------------*/
          if(sRem){/* shift the waveform by sRem bytes */
	  
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	      pWfDest = ((unsigned char *) (regNode->pLocalAddr)) + sRem; 
	      /* this might not work correctly on all platforms */
	      /* especially gig endian processors like power pc could give the wrond readout */

	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned int *) pWfDest)[sDiv]   = ((volatile unsigned int *) pWfTemp)[uiX];
	    }
	    else{/* WF_ADDR_INC */
	      iValTemp =((volatile unsigned int *) regNode->pLocalAddr)[sDiv]  & (0xFFFFFFFF >> ((sizeof(unsigned int) - sRem)<<3));/* mask only the sRem bytes */
	  
	      ((volatile unsigned int *) regNode->pLocalAddr)[sDiv] = iValTemp | (((volatile unsigned int *)pWfTemp)[0]  << ((sRem)<<3));
	      sDiv++;
	      /* pWfTemp=((unsigned char *)pWfTemp)+(sizeof(unsigned int)-sRem); */
	 
	      uiLimit--;
	      for(uiX=0;uiX<uiLimit;uiX++){
		((volatile unsigned int *) regNode->pLocalAddr)[sDiv++] = (((volatile unsigned int *)pWfTemp)[uiX]  >> (( sizeof(unsigned int) - sRem)<<3)) 
		  | (((volatile unsigned int *)pWfTemp)[uiX+1]  << ((sRem)<<3));
	      }
	   
	      iValTemp = ((volatile unsigned int *)regNode->pLocalAddr)[sDiv] & ~((0xFFFFFFFF >> ((sizeof(unsigned int) - sRem)<<3)) );
	      ((volatile unsigned int *) regNode->pLocalAddr)[sDiv] = iValTemp | (((volatile unsigned int *) pWfTemp)[uiX] >> ((sizeof(unsigned int) - sRem)<<3));
	    }

	  } /* sRem != 0 */
          /* ------------------------------ till here ------------------------------------------------------------------ */
	  else{
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned int *) regNode->pLocalAddr)[sDiv]   = ((volatile unsigned int *) pWfTemp)[uiX];  
	    }
	    else{/* WF_ADDR_INC */
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned int *) regNode->pLocalAddr)[sDiv++] = ((volatile unsigned int *) pWfTemp)[uiX];
	    }
	  }
	  
	   
	  PRINT_DEBUG("WF=>  %.8X %.8X %.8X %.8X %.8X \n",((unsigned int*) pWf)[4],
		      ((unsigned int *) pWf)[3],
		      ((unsigned int *) pWf)[2],
		      ((unsigned int *) pWf)[1],
		      ((unsigned int *) pWf)[0]);
	  
	}/* VMED32 */ 
	else if(descrp->sDataMode == VMED16){
	  sDiv=signal>>1; /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned short);

	  PRINT_DEBUG("WF: VMED16 Write data stared at %d offest %d \n",sDiv,sRem);

	  uiLimit=(parm->uiBytesNo)/sizeof(short);
	  PRINT_DEBUG("D16 uiLimit %d\tbytes %d\n",uiLimit,parm->uiBytesNo);
	  pWfTemp=pWf;
          /* -------------- Reminder is different then 0 part 1-------------*/
          if(sRem){/* shift the waveform by sRem bytes */

	      if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
		pWfDest = ((unsigned char *) (regNode->pLocalAddr)) + sRem; 
		/* this might not work correctly on all platforms */
		/* especially gig endian processors like power pc could give the wrond readout */

		for(uiX=0;uiX<uiLimit;uiX++)
		  ((volatile unsigned short *) pWfDest)[sDiv]   = ((volatile unsigned short *) pWfTemp)[uiX];
	      }
	      else{/* WF_ADDR_INC */
		iValTemp =((volatile unsigned short *) regNode->pLocalAddr)[sDiv]  & (0xFFFF >> ((sizeof(unsigned short) - sRem)<<3));/* mask only the sRem bytes */
		
		((volatile unsigned short *) regNode->pLocalAddr)[sDiv] = iValTemp | (((volatile unsigned short *)pWfTemp)[0]  << ((sRem)<<3));
		sDiv++;
		/* pWfTemp=((unsigned char *)pWfTemp)+(sizeof(unsigned short)-sRem); */
	 
		uiLimit--;
		for(uiX=0;uiX<uiLimit;uiX++){
		  ((volatile unsigned short *) regNode->pLocalAddr)[sDiv++] = (((volatile unsigned short *)pWfTemp)[uiX]  >> (( sizeof(unsigned short) - sRem)<<3)) 
		    | (((volatile unsigned short *)pWfTemp)[uiX+1]  << ((sRem)<<3));
		}
		iValTemp = ((volatile unsigned short *)regNode->pLocalAddr)[sDiv] & ~((0xFFFF >> ((sizeof(unsigned short) - sRem)<<3)) );
		((volatile unsigned short *) regNode->pLocalAddr)[sDiv] = iValTemp | (((volatile unsigned short *) pWfTemp)[uiX] >> ((sizeof(unsigned short) - sRem)<<3));
	      }
	  } /* sRem != 0 */
          /* ------------------------------ till here ------------------------------------------------------------------ */
	  else{
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned short *) regNode->pLocalAddr)[sDiv] = ((volatile unsigned short *) pWfTemp)[uiX];
	    }
	    else{/* WF_ADDR_INC */
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned short *) regNode->pLocalAddr)[sDiv++] = ((volatile unsigned short *) pWfTemp)[uiX];
	    }
	  }
	  
	  PRINT_DEBUG("WF=>  %.4X %.4X %.4X %.4X\n",((unsigned short*) pWf)[3],
		      ((unsigned short *) pWf)[2],
		      ((unsigned short *) pWf)[1],
		      ((unsigned short *) pWf)[0]);
	  
	  
	}/* VMED16 */ 


	else if(descrp->sDataMode == VMED8){
	  sDiv=signal;
	  PRINT_DEBUG("D08 uiLimit %d\tbytes %d\n",parm->uiBytesNo,parm->uiBytesNo);
	  if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	    for(uiX=0;uiX<parm->uiBytesNo;uiX++){
	      ((volatile unsigned char *) regNode->pLocalAddr)[sDiv]   = ((volatile char *)pWf)[uiX];
	    }
	  }
	  else{/* WF_ADDR_INC */
	    for(uiX=0;uiX<parm->uiBytesNo;uiX++){
	      ((volatile unsigned char *) regNode->pLocalAddr)[sDiv++] = ((volatile char *)pWf)[uiX];
	    }
	  }
	  PRINT_DEBUG("WF=>  %.8X\n",((unsigned int*) pWf)[0]);


	}/* VMED8 */
	}/* (status & ENTER_FUNC) == RETURN_OK */

	if((status & ENTER_FUNC) == RETURN_OK){
	  if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    status = regNode->fnct(card,signal,parm,REC_WF,REC_EXIT,chanName,(void *)pWf);
	  }	  
	}/* (status & ENTER_FUNC) == RETURN_OK */


	if((status & ERROR_OCCURED) == RETURN_OK){
	  PRINT_DEBUG("Before leaving WF_GPMM \n");
	  
	}
	else
	  {
	    printf("An error NR 0x%.4X  occured in WF:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }
	  
      }/* regNode != NULL */

    return(Ok);

}



/*------------------------------------------------------*/
/*               WF   read                              */
/* cShiftBits refers to the number of bytes per element */
/* ex: short -> 2 bytes                                 */
/*------------------------------------------------------*/

int READ_WF_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    char  *chanName,
    void  *pWf
    ){

    struct io_GPMM	  *descrp=NULL;
    struct register_Node  *regNode=NULL;
    int                    status = RETURN_OK;
    unsigned short         sRem=0,sDiv=0; /* Reminder, Devision */
    /* volatile unsigned int   iVal = 0; */
    unsigned int           uiX=0;
    unsigned int           uiLimit=0;
    unsigned int           uiFlagTmp=0;

#ifdef DEBUG_SOFT    
    char           lTmpTab[40]="ALA ma Asa a Ola ma kota."; /* TEST ONLY this line should be deleted */
#endif

    PRINT_DEBUG("READ_WF_GPMM Chan name %s #C%d S%d @%s\n",chanName,card,signal,parm->sDesc);
    if(pWf == NULL){
      printf("WF: wf points to NULL\n");
      return (-1);
    }

    if(parm!=NULL)
      regNode=parm->pRegLocation;
 
    /* if the desired record was found then set the value according to the signal number "iRefNr" */
      if( regNode != NULL ){
	if((descrp=regNode->pCard)==NULL){
	  printf("No card associated with this register Node %s\n",regNode->sDesc);
	  return (-1);
	}
#ifdef DEBUG_SOFT
	regNode->pLocalAddr=lTmpTab; /* TEST ONLY this line should be deleted */
#endif
	

	if(regNode->fnct != NULL){
	  /* execut the user defined process function */
	    /* the manipulation is done on the lVal what coresponds to the rval field */  
	   status = regNode->fnct(card,signal,parm,REC_WF,REC_ENTER,chanName,(void *)pWf);
	}

	/* this is the way how to access data in D32 mode i.e. intiger by integer */
	if ( (status & PROCESS_DATA) == RETURN_OK){
	if(descrp->sDataMode == VMED32){
	  sDiv=signal>>2; /* signal/sizeof(unsigned int)=signal>>2 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned int);

	  PRINT_DEBUG("WF: VMED32 Read data  stared at %d offest %d \n",sDiv,sRem);

	  uiLimit=(parm->uiBytesNo)/sizeof(int);
	  PRINT_DEBUG("D32 uiLimit %d\tbytes %d\n",uiLimit,parm->uiBytesNo);
          /* -------------- Reminder is different then 0 part 1-------------*/
          if(sRem){/* shift the waveform by sRem bytes */	 
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ)
	      uiFlagTmp=0;
	    else
	      uiFlagTmp=1;
	    for(uiX=0;uiX<uiLimit;uiX++){
	    ((volatile unsigned int *)pWf)[uiX] = (((volatile unsigned int *) regNode->pLocalAddr)[sDiv]  >> ((sRem)<<3)) 
	      | (((volatile unsigned int *) regNode->pLocalAddr)[sDiv+1] << (( sizeof(unsigned int) - sRem)<<3));
	    sDiv=sDiv+uiFlagTmp;
	    }


	  } /* sRem != 0 */
          /* ------------------------------ till here ------------------------------------------------------------------ */
	  else{
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned int *) pWf)[uiX] = ((volatile unsigned int *) regNode->pLocalAddr)[sDiv];
	    }
	    else{/* WF_ADDR_INC */
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned int *) pWf)[uiX] = ((volatile unsigned int *) regNode->pLocalAddr)[sDiv++];
	    }
	  }
	  
	  PRINT_DEBUG("WF=>  %.8X %.8X %.8X %.8X %.8X \n",((unsigned int*) pWf)[4],
		      ((unsigned int *) pWf)[3],
		      ((unsigned int *) pWf)[2],
		      ((unsigned int *) pWf)[1],
		      ((unsigned int *) pWf)[0]);
	  
	}/* VMED32 */ 
	else if(descrp->sDataMode == VMED16){
	  sDiv=signal>>1; /* signal/sizeof(unsigned short)=signal>>1 ; to speed up the operation */
	  sRem=signal%sizeof(unsigned short);

	  PRINT_DEBUG("WF: VMED16 Read data  stared at %d offest %d \n",sDiv,sRem);

	  uiLimit=(parm->uiBytesNo)/sizeof(short);
	  PRINT_DEBUG("D16 uiLimit %d\tbytes %d\n",uiLimit,parm->uiBytesNo);
          /* -------------- Reminder is different then 0 part 1-------------*/
          if(sRem){/* shift the waveform by sRem bytes */	 
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ)
	      uiFlagTmp=0;
	    else
	      uiFlagTmp=1;
	    for(uiX=0;uiX<uiLimit;uiX++){
	    ((volatile unsigned short *)pWf)[uiX] = (((volatile unsigned short *) regNode->pLocalAddr)[sDiv]  >> ((sRem)<<3)) 
	      | (((volatile unsigned short *) regNode->pLocalAddr)[sDiv+1] << (( sizeof(unsigned short) - sRem)<<3));
	    sDiv=sDiv+uiFlagTmp;
	    }

	  } /* sRem != 0 */
          /* ------------------------------ till here ------------------------------------------------------------------ */
	  else{
	    if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned short *) pWf)[uiX] = ((volatile unsigned short *) regNode->pLocalAddr)[sDiv];
	    }
	    else{/* WF_ADDR_INC */
	      for(uiX=0;uiX<uiLimit;uiX++)
		((volatile unsigned short *) pWf)[uiX] = ((volatile unsigned short *) regNode->pLocalAddr)[sDiv++];
	    }
	  }
	  
	  PRINT_DEBUG("WF=>  %.4X %.4X %.4X %.4X\n",((unsigned short*) pWf)[3],
		      ((unsigned short *) pWf)[2],
		      ((unsigned short *) pWf)[1],
		      ((unsigned short *) pWf)[0]);
	  
	}/* VMED16 */ 


	else if(descrp->sDataMode == VMED8){
	  sDiv=signal;
	  PRINT_DEBUG("D08 uiLimit %d\tbytes %d\n",parm->uiBytesNo,parm->uiBytesNo);
	  if ((parm->uiParam1 & WF_ACCESS_TYPE) == WF_SEQ){
	    for(uiX=0;uiX<parm->uiBytesNo;uiX++){
	      ((volatile char *)pWf)[uiX] = ((volatile unsigned char *) regNode->pLocalAddr)[sDiv];
	    }
	  }
	  else{/* WF_ADDR_INC */
	    for(uiX=0;uiX<parm->uiBytesNo;uiX++){
	      ((volatile char *)pWf)[uiX] = ((volatile unsigned char *) regNode->pLocalAddr)[sDiv++];
	    }
	  }
	  PRINT_DEBUG("WF=>  %.8X\n",((unsigned int*) pWf)[0]);


	}/* VMED8 */
	}/* (status & ENTER_FUNC) == RETURN_OK */

	if((status & ENTER_FUNC) == RETURN_OK){
	  if(regNode->fnct != NULL){
	    /* execut the user defined process function */
	    status = regNode->fnct(card,signal,parm,REC_WF,REC_EXIT,chanName,(void *)pWf);
	  }	  
	}/* (status & ENTER_FUNC) == RETURN_OK */


	if((status & ERROR_OCCURED) == RETURN_OK){
	  PRINT_DEBUG("Before leaving WF_GPMM \n");
	  
	}
	else
	  {
	    printf("An error NR 0x%.4X  occured in WF:\n #C%d S%d @%s\n",status,card,signal,parm->sDesc);
	    return(-1);
	  }
	  
      }/* regNode != NULL */

    return(Ok);

}

/*----------------------------------------------------*/
/* IO_REPORT prints the report in the minicom window  */
/*----------------------------------------------------*/

static long IO_REPORT(int iReportLevel)
{
    struct devInfo        *pDevBlock;
    struct io_GPMM	  *pPrivate;
    int 		   noCard;
    char                  sStr[10];
    short                 ix=0;
    struct register_Node  *regNode;
    /*    int                    iIndex;*/
    pDevBlock = slsDevFirst( pGPMMList );
    printf("Type:\n");
    printf("dbior \"drvGPMM\",1 => Cards config info\n");
    printf("dbior \"drvGPMM\",2 => Register redouts\n\n");
    
    
    if(iReportLevel>=1 && iReportLevel<=2){
    noCard=0;
    while (pDevBlock != NULL) {
      pPrivate = slsDevPrivate(pDevBlock);
       printf("======== Card nr %d %-10.10s %.3s %.3s =======\n",(int) pDevBlock->cardNum,
	      pPrivate->pcBoardName,
	      (pPrivate->sAddrMode==atVMEA32)? "A32":((pPrivate->sAddrMode==atVMEA24)? "A24":"A16"),
	      (pPrivate->sDataMode==VMED32)?   "D32":((pPrivate->sDataMode==VMED16)? "D16":"D8"));

      /*---------  List of registers ------------*/
      if (iReportLevel == 1){
#if (defined(EPICS_REVISION) && (EPICS_REVISION == 14))
      printf("Base address 0X%x Registers:\n",slsDevAddrs(pDevBlock));
#else
      printf("Base address %p Registers:\n",slsDevAddrs(pDevBlock));
#endif
      printf("Nr\tAddr offset\tName       configured\n");
      regNode=pPrivate->pRegisterNode;
      while( regNode != NULL){

	printf("%d\t0x%x\t\t%-10.10s %s\n",regNode->iRefNr,
	       regNode->iAddrOffset,
	       regNode->sDesc,
	       (regNode->sConfigured==TRUE)? "Y":"N");
	 regNode = regNode->nextNode;
      }

#ifndef linux
      if(pPrivate->paioscanpvt){
	printf("-------- Interrupt Handler Registered -------\n");
	printf("Vector\tLevel\t\tISR execution count\n");
	printf("0x%.2X\t0x%.2X\t\t0x%.8X\n",(unsigned int)pDevBlock->vectorNum,(unsigned int)pDevBlock->itrLevel,pPrivate->iIntrCounter);
      }
#endif
      
      }

      /*---------  Register conetents ------------*/
      else if(iReportLevel ==2){
 
	printf("Byte nr:      6   4    2   0\n");
	printf("Register:\n");
	
	regNode=pPrivate->pRegisterNode;
      while( regNode != NULL){

	printf("%-10.10s ",regNode->sDesc);
	if(pPrivate->sDataMode == VMED32){


	  for (ix=2;ix>0;ix--){
	    
	    if((short)(((regNode->uiBytesNr)+3)/4)>=ix){
	      sprintf(sStr,"%.8X",((unsigned int *) regNode->pLocalAddr)[ix-1]);	      
	      printf("%c%c%c%c%c%c%c%c ",sStr[0],sStr[1],sStr[2],sStr[3],sStr[4],sStr[5],sStr[6],sStr[7]); 
	    }
	    else
	      printf("         ");

	  }

	} /* VMED32 */
		if(pPrivate->sDataMode == VMED16){

	  for (ix=4;ix>0;ix--){
	    if(ix==2)
		printf(" ");
	    if((short)(((regNode->uiBytesNr)+1)/2)>=ix){
	      sprintf(sStr,"%.4X",((unsigned short *) regNode->pLocalAddr)[ix-1]);	      
	      printf("%c%c%c%c",sStr[0],sStr[1],sStr[2],sStr[3]); 
	    }
	    else
	      printf("    ");

	  }


	} /* VMED16 */
	  else if(pPrivate->sDataMode == VMED8){
	  for (ix=7;ix>=0;ix--){
	    if(ix==3)
		printf(" ");
	    if(((short)(regNode->uiBytesNr))>ix){
	      sprintf(sStr,"%.2X",((unsigned char *) regNode->pLocalAddr)[ix]);	      
	      printf("%c%c",sStr[0],sStr[1]);
	    }
	    else
	      printf("  ");

	  }
	  }
	printf("\n");


	 regNode = regNode->nextNode;
      }


      }
      

      printf("\n");
      noCard++;
      pDevBlock=slsDevNext(pDevBlock);
    }

#ifdef linux
     if (iReportLevel == 1){
	printf("-------- Interrupt Handler Registered -------\n");
	printf("IRQ NR\t\tISR execution count\n");
	for(ix=0;ix<IRQ_NR_LIMIT;ix++)
	  if(irqGpmmSup.pioScanPvt[ix])
	    printf("0x%.2X\t\t0x%.8X\n",ix,(unsigned int)(irqGpmmSup.piIrqCount[ix]));
     }
#endif

    printf("=============================================\n");
    }
    return(Ok);
}

/*-------------------------------------------------------*/
/* functiones called from the device support             */
/*-------------------------------------------------------*/

int getRegLocation(unsigned short sCardNr, char *parm,void **pRegNode){

  struct io_GPMM	        *descrp;
  struct register_Node  *regNode;
  char                  pcStr[100];
  char                 *pCh=NULL;

  /* search for the given card number "card" */

    if ( PARAMS_NOT_VALID(pGPMMList,descrp,sCardNr) ) {
      if (descrp == NULL) return S_drvGPMM_noDevice;
      if (!descrp->configured) return S_drvGPMM_noDevice;
      return S_drvGPMM_badParam;
    }
    /* search for the given record referenced by the "parm" */
    regNode=descrp->pRegisterNode;

    sprintf(pcStr,"%.70s",parm);
    /* extract the register name form the parametre name */
    pCh=strchr(pcStr,':'); /* search for the first occurence of : which separates the user def Function input params */    
    if(pCh!=NULL)
      pCh[0]='\0';
    else{
      pCh=strchr(pcStr,','); /* search for the first occurence of comma which separates epics channel arguments */
    if(pCh!=NULL)
      pCh[0]='\0';
    }
      while( regNode != NULL){
	if (! strcmp(regNode->sDesc,pcStr))
	  break;
	 regNode = regNode->nextNode;
      }
      
      if ( regNode != NULL ){
	*pRegNode=(void *)regNode;
	return Ok;
      }
      else
	return S_drvGPMM_noDevice;
      
}


/* ------------------------------ test user defined functions ---------------------*/

/* example for MBBODirect record */
int testGPMMFFuncAdc(
    unsigned short card,
    unsigned short signal,
    struct recDesc   *parm,
    unsigned short   sRecordType,
    char *chanName,
    void *pVal
)
{
  unsigned short sVal;
  sVal=*((unsigned short *)pVal);
  printf("C%d S%d @%s testFunc %.8X\n",card,signal,parm->sDesc,sVal);
  
  if(sVal>0x100)
    return (Ok);  /* sucess */
  else
    return (-1);  /* error occured */
}


/* example for MBBIDirect record */

int testGPMMFFilterHiByte(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short   sRecordType,
    unsigned short   sRecordPass,
    char *chanName,
    void *pVal
)
{
  short sVal;
  
  sVal=*((unsigned short *)pVal);
  printf("Function testFilterHiByte C%d S%d @%s %.8X\n",card,signal,parm->sDesc,sVal);
  sVal=sVal & 0x00FF;
  printf("Function testFilterHiByte C%d S%d @%s %.8X\n",card,signal,parm->sDesc,sVal);

  *((unsigned short *)pVal)=sVal;
  return (Ok);

}

/* example of the general purpose debug function record */


int debugGPMMFunc(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short   sRecordType,
    unsigned short   sRecordPass,
    char *chanName,
    void *pVal
)
{
  unsigned long  lVal;
  unsigned short sVal;
  char *pPropertyName;
  short status=RETURN_OK;

  lVal= *((unsigned long  *)pVal);   /* for ai ao record. lVal refers to the rval field */
  sVal= *((unsigned short *)pVal);   /* for mbbi mbbo bi bo */
  
  pPropertyName=strchr(chanName,':');
  
  /*---------------------------------*/
  /* code executed on function enter */
  /*---------------------------------*/
  if(sRecordPass==REC_ENTER){
    /* do the data processing when enetring the driver writ/read function for the specyfic record */
    printf("FUNCTION PASS on ENTER\n");
    if(sRecordType==REC_AI){
      printf("Record AI\n");
      printf("Function debugGPMMFunc C%d S%d @%s %s 0x%.8X\n",card,signal,
	     parm->sDesc,chanName,(unsigned int) lVal);
    }
    else if(sRecordType==REC_AO){
      printf("Record AO\n");
      printf("Function debugGPMMFunc C%d S%d @%s %s 0x%.8X\n",card,signal,
	     parm->sDesc,chanName,(unsigned int) lVal);
    }
    else if(sRecordType==REC_MBBI_D){
      printf("Record MBBI DIRECT\n");
      printf("Function debugGPMMFunc C%d S%d @%s %s 0x%.4X\n",card,signal,parm->sDesc,
	     chanName,(unsigned short) sVal);
    }
    else if(sRecordType==REC_MBBO_D){
      printf("Record MBBO DIRECT\n");
      printf("Function debugGPMMFunc C%d S%d @%s %s 0x%.4X\n",card,signal,parm->sDesc,
	     chanName,(unsigned short) sVal);
    }
    else if(sRecordType==REC_BI){
      printf("Record BI \n");
      printf("Function debugGPMMFunc C%d S%d @%s %s 0x%.4X\n",card,signal,parm->sDesc,
	     chanName,(unsigned short) sVal);
    }
    else if(sRecordType==REC_BO){
      printf("Record BO \n");
      printf("Function debugGPMMFunc C%d S%d @%s %s 0x%.4X\n",card,signal,parm->sDesc,
	     chanName,(unsigned short) sVal);
    }

  }
  /*--------------------------------*/
  /* code executed on function exit */
  /*--------------------------------*/
  else{
    printf("FUNCTION PASS on EXIT\n");
  }
   
  return (status);

}

/* -------------------- end of user defined functions ----------- */


int debug2GPMMFunc(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short   sRecordType,
    unsigned short   sRecordPass,
    char *chanName,
    void *pVal
)
{
  unsigned long  lVal;
  unsigned short sVal;
  char *pPropertyName;
  short status=RETURN_OK;

  struct io_GPMM	  *pPrivate   =NULL;
  struct register_Node    *pRegNode   =NULL;


  lVal= *((unsigned long  *)pVal);   /* for ai ao record. lVal refers to the rval field */
  sVal= *((unsigned short *)pVal);   /* for mbbi mbbo bi bo */
  
  pPropertyName=strchr(chanName,':');

  /*---------------------------------*/
  /* code executed on function enter */
  /*---------------------------------*/
  if(sRecordPass==REC_ENTER){
    
    /* card related data */
    if(parm!=NULL){
      pRegNode=parm->pRegLocation;
      if( pRegNode != NULL ){
	pPrivate=pRegNode->pCard;

	printf("============= Board related data ===========\n");
    
	printf("Board %d Name %-10.10s\t %.3s\t %.3s\n\n",
	     card,pPrivate->pcBoardName,
	     (pPrivate->sAddrMode==atVMEA32)? "A32":((pPrivate->sAddrMode==atVMEA24)? "A24":"A16"),
	     (pPrivate->sDataMode==VMED32)?   "D32":((pPrivate->sDataMode==VMED16)? "D16":"D8"));
#if (defined(EPICS_REVISION) && (EPICS_REVISION == 14))
	printf("\tBase Addr OX%x Intr Vector 0X%.2x Level %d\n",
#else
        printf("\tBase Addr %p Intr Vector 0X%.2x Level %d\n",
#endif
	       slsDevAddrs(getGPMMDevBlock(card)),
	       (unsigned int) slsDevVectorNum(getGPMMDevBlock(card)),
	       (unsigned int)slsDevItrLevel(getGPMMDevBlock(card)));

        printf("EPICS channel related data:\n");
	printf("Channel name: %s\n",chanName);
	printf("Signal Nr\tParam\n");
	printf("%d\t%s\n",signal,parm->sDesc);
        printf("Register Ref Nr\tAddr offset\tName       configured\n");
	printf("%d\t0x%x\t\t%-10.10s %s\n",pRegNode->iRefNr,
	       pRegNode->iAddrOffset,
	       pRegNode->sDesc,
	       (pRegNode->sConfigured==TRUE)? "Y":"N");
	printf("Number of mask bits    | bit mask | Bytes number | Shift bits\n");
	printf("%d\t\t\t0x%.8X\t %d\t%d\n",
	       parm->cMaskBitNr,
	       parm->uiMask,
	       parm->uiBytesNo,
	       parm->cShiftBits);
        printf("User function arguments\n");
        printf("Input Arg 1 | Input Arg 2 | Output Arg\n");
	printf("%d\t%d\t%p\n",
	       (int) parm->lInArg1,
	       (int) parm->lInArg2,
	       parm->vOutArg);
        printf("Memory location to access data %p\n",pRegNode->pLocalAddr);
	printf("Channel RVAL (short) 0x%.4X (long) 0x%.8X\n",sVal,(int)lVal);
    
	  }
    }

  }
  /*--------------------------------*/
  /* code executed on function exit */
  /*--------------------------------*/
  else{
    printf("FUNCTION PASS on EXIT\n");
  }
   
  return (status);

}
