#include <dbScan.h>
#define Ok 0
#define ERR -1
#define EOS '\0'

/* data access mode on VME bus */
#define VMED8  0
#define VMED16 1
#define VMED32 2

#define STATIC static

/*#define DEBUG          This is uesd to print debug messges  */
/*#define DEBUG_SOFT   this emulates hardware in the VME   */

/* the alternative for the
PRINT_DEBUG(args...) printf(args)
could be
PRINT_DEBUG(...) printf(__VA_ARGS__)
*/

#ifdef DEBUG
#define PRINT_DEBUG(args...) printf("DEBUG: "); printf(args)
#else
#define PRINT_DEBUG(args...)
#endif

struct io_GPMM {
           char      pcBoardName[16];                /* board name */
           unsigned short sAddrMode;                 /* VME Address access mode */
           unsigned short sDataMode;                 /* VME data access mode */
           struct register_Node  *pRegisterNode;     /* link list of board registers */                        
           IOSCANPVT      paioscanpvt;               /* I/O Event Scanning */
           unsigned int   iIntrCounter;              /* Interrupts counter */

    int configured;
  /* int enableItrOnInit; */
};

#define WF_REC_IN      0
#define WF_REC_OUT     1
#define WF_REC_TYPE    0x1

#define WF_SEQ         0x2           /* sequential access to the device memory location          */
#define WF_ADDR_INC    0x0           /* address incremental access to the device memory location */  
#define WF_ACCESS_TYPE 0x2


struct recDesc {
  char                 *sDesc; /* [51] */
  void                 *pRegLocation;      /* pointer to the associated register memory 
					      location allocated in the driver supprot          */
  unsigned char        cMaskBitNr;          /* Number of mask bits                              */
  unsigned int         uiMask;              /* bit mask for AI AO MBBI MBBO records.            */
  unsigned int         uiBytesNo;            /* Number of bytes to consider: AI, AO, 
					       String In records.                                */
  
  unsigned char        cShiftBits;          /* Number of bits to shift    for MBBI MBBO BI BO   */
  unsigned int         uiParam1;             /* this defines if the WF  record is IN or OUT */
           int         iIntrNr;             /* interrupt number */
  /* input and output arguments for the use in the user define function */
  long lInArg1;                             /* input arguments */
  long lInArg2;
  void *vOutArg;                            /* output argument */
};

struct register_Node {
  unsigned short        iRefNr;      /* as the signal number in INP field */
  unsigned int          iAddrOffset;
  void                 *pLocalAddr;  /* local addres of the register */
  char                  sDesc[51];   /* register name */
  unsigned int          uiBytesNr;    /* register size in bytes */
                                     /* userdefined function. return Ok or -1 */
  char                  cRW;         /* data to be read out (R) written (W) ignore (X) */
  int                   (*fnct)(unsigned short, unsigned short,struct recDesc *, unsigned short, unsigned short, char *, void *);
  /* short                 sAddrMod;     A16 or A24 or A32 */
  void                 *pCard;       /* pointer to the card associated with the given register node */

  short                 sConfigured; /* register properly mapped 0 No 1 Yes */ 
  struct  register_Node *nextNode;
};

#ifndef linux
void addGPMMInterrupt( int iCardNr, int iIntLevel, int iVector, void (*gpmmUserISR)(struct io_GPMM *pPrivate));
#else
void addGPMMInterrupt( int iCardNr, int iIntrNr, unsigned int uiBaseAddr, void (*gpmmUserISR)(struct io_GPMM *pPrivate));
#endif

int addGPMMRegister(
	     unsigned int         cardNum,        /* card number as used in INP fields */
	     unsigned short       iRefNr,         /* signal number as used in INP field */
	     unsigned int         iAddrOffset,    /* ofset from the base addres */
	     unsigned int       uiBytesNr,       /* size of the register in Bytes */
	     char                 cRW,            /* data to be read (R) written (W) or nothing (X) */
	     char                *sDesc,          /* register description 20 char's */
             int                 (*fnct)(unsigned short, unsigned short, struct recDesc *, unsigned short, unsigned short, char *, void *)
	     );

void GPMMConfigure(
    int         cardNum,          /* card number as used in INP fields */
    void        *base,            /* base address of the board */
    char        *pcBoardName,      /* name of the board */
    char        *sAddrAccessMode, /* A16 A24 A32 */
    char        *sDataAccessMode  /* D8 D16 D32 */			
    );

int debugGPMMFunc(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short   sRecordType,
    unsigned short   sRecordPass,
    char *chanName,
    void *pVal
    );

int debug2GPMMFunc(
    unsigned short card,
    unsigned short signal,
    struct recDesc  *parm,
    unsigned short   sRecordType,
    unsigned short   sRecordPass,
    char *chanName,
    void *pVal
    );


#define READ_MBBI_DIRRECT_GPMM   GPMM_read_MBBI_Dirrect
#define WRITE_MBBO_DIRRECT_GPMM  GPMM_write_MBBO_Dirrect
#define READ_AI_GPMM             GPMM_read_AI
#define WRITE_AO_GPMM            GPMM_write_AO
#define WRITE_WF_GPMM            GPMM_write_WF


int READ_AI_GPMM(
    unsigned short   card,
    unsigned short   signal,
    struct recDesc  *param,
    unsigned short   sRecordType,
    char            *chanName,
    unsigned int    *val

);	

int WRITE_AO_GPMM(
    unsigned short   card,
    unsigned short   signal,
    struct recDesc  *param,
    unsigned short   sRecordType,
    char            *chanName,
    unsigned int     val
);

int READ_BI_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc *param,
    char  *chanName,
    unsigned short *val
);

int WRITE_BO_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc *param,
    char  *chanName,
    unsigned short val
);

int READ_MBBI_DIRRECT_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc *param,
    unsigned short  sRecordType,
    char  *chanName,
    unsigned short *val
);

int WRITE_MBBO_DIRRECT_GPMM(
    unsigned short card,
    unsigned short mask,
    struct recDesc  *parm,
    unsigned short   sRecordType,
    char  *chanName,
    unsigned short val
);

int READ_STR_IN_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc *param,
    char  *chanName,
    char  *val
);

int READ_WF_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc *param,
    char  *chanName,
    void  *val
);

int WRITE_WF_GPMM(
    unsigned short card,
    unsigned short signal,
    struct recDesc *param,
    char  *chanName,
    void  *val
);

/* function used to obtaine the pointer to the regiser node referenced by the card number and the parameter string */ 
int getRegLocation(unsigned short sCardNr, char *parm, void **pRegNode);
struct devInfo *getGPMMDevBlock(int iCardNumber);



#define M_drvGPMMLib (1003<<16U)

#define drvGPMMError(CODE) (M_drvGPMMLib | (CODE))

#define S_drvGPMM_OK 0 /* success */
#define S_drvGPMM_badParam drvGPMMError(1) /*GPMM driver: bad parameter*/
#define S_drvGPMM_noMemory drvGPMMError(2) /*GPMM driver: no memory*/
#define S_drvGPMM_noDevice drvGPMMError(3) /*GPMM driver: device not configured*/
#define S_drvGPMM_invSigMode drvGPMMError(4)/*GPMM driver: signal mode conflicts with device config*/
#define S_drvGPMM_cbackChg drvGPMMError(5) /*GPMM driver: specified callback differs from previous config*/
#define S_drvGPMM_alreadyQd drvGPMMError(6)/*GPMM driver: a read request is already queued for the channel*/
#define S_drvGPMM_null drvGPMMError(7) /* GPMM local Addr == Null */


#define INIT_FROM_HARDWARE  1
#define INIT_RESTORE_VAL    2



#ifdef linux
#define IRQ_NR_LIMIT       32 /* number of interrupts [0,IRQ_NR_LIMIT-1] */
struct sIrqGpmmSup{
  IOSCANPVT pioScanPvt[IRQ_NR_LIMIT];
  unsigned int piIrqCount[IRQ_NR_LIMIT];
  unsigned int irqHandlerOn;
  int      gpmmDrvFD; /* file descriptor */
};
#endif
