/*start of CVS identification*/
/*$Author: zimoch $*/
/*$Date: 2006/12/21 14:27:56 $*/
/*$Id: slsDevLib.c,v 1.2 2006/12/21 14:27:56 zimoch Exp $*/
/*$Name:  $*/
/*$Revision: 1.2 $*/
/*end of CVS identification*/

#ifndef __GNUC__
#define __attribute__(x)
#endif

static char rcsid[] __attribute__((unused)) =
    "$Id: slsDevLib.c,v 1.2 2006/12/21 14:27:56 zimoch Exp $";
 
/*
 *	KECK Observatory device library.
 *		
 *
 * 	Author:		Allan Honey
 * 	Date:		1-30-96
 *
 */
#ifndef linux
#include <vxWorks.h>
#include <vxLib.h>
#include <sysLib.h>
#include <logLib.h>
#endif /* linux */

#include <stdio.h>
#include <stdlib.h>

#include <slsDevLib.h>



void slsDevConfigure(
    char	*devName,
    long	cardNum,
    void	*baseAddrs,
    long	vectorNum,
    long 	itrLevel,
    void	*devPrivate,


    struct devInfo 	**pInfoBlock
)
{
    struct devInfo	*pInfo;

    /*
     * allocate a structure for the board info.
     */
    pInfo = calloc( 1, sizeof(struct devInfo));
    if ( pInfo == NULL ) {
#ifndef linux
	logMsg ("%s: memory allocation error. Board not configured\n", 
		(int)devName,0,0,0,0,0 );
#else
	fprintf(stderr,"slsDevLib: memory allocation error. Board not configured\n");
#endif
	return;
    }

    pInfo->devAddrs = baseAddrs;
    pInfo->cardNum = cardNum;
    pInfo->vectorNum = vectorNum;
    pInfo->itrLevel = itrLevel;
    pInfo->devPrivate = devPrivate;
    pInfo->nextCard = *pInfoBlock;
    *pInfoBlock = pInfo;
}

long slsDevInUse(
    void	*baseAddrs,
    long	cardNum,
   
    struct devInfo	**pInfoBlock
)
{
    struct devInfo	*pInfo;

    if (*pInfoBlock == NULL) return 0;

    pInfo = *pInfoBlock;
    while (pInfo != NULL) {
	if ( (pInfo->devAddrs == baseAddrs) || (pInfo->cardNum == cardNum) ) 
	    return -1;
	pInfo = pInfo->nextCard;
    }

    return 0;
}

void slsDevList(
    char	    *devType,
    struct devInfo  *pDevList
)
{
    struct devInfo *pBoard;
    
    pBoard = pDevList;
    printf("\n%s:\n",devType);
    printf(" card    device   vector  interrupt\n");
    printf("number   address  number    level  \n");

    while (pBoard != NULL) {
	printf("  %2ld    %8x    %2ld       %2ld\n",
	    pBoard->cardNum, (unsigned int)pBoard->devAddrs, 
	    pBoard->vectorNum, pBoard->itrLevel );

	pBoard = pBoard->nextCard;
    }

    printf("\n");

} 

struct devInfo *slsDevNodeByCardNum(
    struct devInfo 	*devList,
    long    		cardNum
)
{
    struct devInfo	*node;

    node = slsDevFirst(devList);

    while (node != NULL) {
	if (node->cardNum == cardNum) break;

	node = slsDevNext(node);
    }

    return node;
}

long slsDevChgBase( 
    struct devInfo	*node,
    void		*baseAddrs
)
{
    if (node == NULL) return 1;

    node->devAddrs = baseAddrs;

    return 0;	
}

/*+*********************************************************************
  $Log: slsDevLib.c,v $
  Revision 1.2  2006/12/21 14:27:56  zimoch
  fixes for 3.14

  Revision 1.1  1999/08/26 12:18:56  hunt
  *** empty log message ***

  Revision 1.20  1999/08/24 14:49:19  hunt
  Added make release to makefile

  Revision 1.19  1999/08/24 14:20:33  hunt
  Added cvsId as static char[]

  Revision 1.18  1999/08/23 07:36:27  hunt
  Fixed bug in MBBO mask was 0 for bits >16

  Revision 1.17  1999/08/18 09:20:09  hunt
  Fixed problem with MBBO - incorrect init value to csrIO (1000 not 0c00)

  Revision 1.16  1999/08/17 16:28:02  hunt
  Driver changed to invert sense of output for 8303 (current output)
  transition module - it now matches the 8301 input module

  Revision 1.15  1999/08/17 14:54:15  hunt
  Hot Swap now works under interrupt (I/O Intr)

  Revision 1.13  1999/08/09 16:29:32  hunt
  Some problems with multiple input cards

  Revision 1.12  1999/08/05 12:02:31  hunt
  Input card only works when first IO card in system

  Revision 1.11  1999/08/04 18:43:01  hunt
  I/O Intr now seems to work for BI

  Revision 1.10  1999/08/04 18:14:43  hunt
  IRQ working only with single card

  Revision 1.9  1999/08/04 11:30:04  hunt
  Added interrup support

  Revision 1.8  1999/08/03 11:27:33  hunt
  MBBO works lower and upper 32bit chans.

  Revision 1.7  1999/08/02 15:51:12  hunt
  MBBI now works with inputs >32 - use "1" as param in .db to select

  Revision 1.1  1999/07/09 10:56:48  hunt
  *** empty log message ***

  Revision 1.7  1999/06/15 17:45:12  hunt
  now uses 'generic' driver template with ai,ao,bo.bi,mbbi,mbbo

  Revision 1.1.1.1  1999/05/25 14:40:30  higgs
  Original version

  Revision 1.3  1998/03/11 02:15:42  ahoney
  Modified printf statement to eliminate compiler warnings.

  Revision 1.2  1996/05/04 01:51:25  ahoney
  Added koDevChgBase() to provide for dynamic changing of a card's base
  address. This facilitates boards checkout/diagnostics.

 * Revision 1.1  1996/02/13  21:58:59  ahoney
 * Initial addition for directory koDevLib (related to module_types
 * elimination). This module contains defines and routines relevant to
 * maintaining dev info lists.
 *
 *
***********************************************************************/

